from typing import TYPE_CHECKING

from bgpsimulator.shared.enums import Relationships

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class ASPathEdgeFilter:
    """A Policy that filters announcements based on the edge of the AS-Path"""

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Returns invalid if an edge AS is announcing a path containing other ASNs"""

        origin_asn = ann.as_path[0]

        if origin_asn in policy.as_.neighbor_asns:
            neighbor_as_obj = policy.as_.as_graph.as_dict[origin_asn]
            if (neighbor_as_obj.stub or neighbor_as_obj.multihomed) and set(
                ann.as_path
            ) != {neighbor_as_obj.asn}:
                return False
        return True



import json
from pathlib import Path

from bgpsimulator.shared import Outcomes
from bgpsimulator.simulation_engine import SimulationEngine
from bgpsimulator.simulation_framework import DataPlanePacketPropagator, Scenario

from .diagram import Diagram
from .engine_run_config import EngineRunConfig


class EngineRunner:
    """Runs a single engine run"""

    def __init__(
        self,
        engine_run_config: EngineRunConfig,
        base_dir: Path = Path.home() / "Desktop" / "bgpsimulator_engine_runs",
        overwrite: bool = False,
        compare_against_ground_truth: bool = False,
        write_diagrams: bool = True,
    ):
        self.conf: EngineRunConfig = engine_run_config
        self.base_dir: Path = base_dir
        self.overwrite: bool = overwrite
        # True when used for tests, False when used for everything else
        self.compare_against_ground_truth: bool = compare_against_ground_truth
        self.storage_dir: Path = self.base_dir / self.conf.name
        self.storage_dir.mkdir(parents=True, exist_ok=True)
        self.write_diagrams: bool = write_diagrams

    def run(self, dpi: int | None = None):
        """Runs the engine run"""

        engine, scenario = self._get_engine_and_scenario()

        # Run engine
        for round_ in range(self.conf.scenario_config.propagation_rounds):
            engine.propagate(propagation_round=round_, scenario=scenario)
            # By default, these are both no ops
            for func in (scenario.pre_aggregation_hook, scenario.post_propagation_hook):
                func(
                    engine=engine,
                    propagation_round=round_,
                    trial=0,
                    percent_ases_randomly_adopting=0,
                )
        data_plane_packet_propagator = DataPlanePacketPropagator()
        data_plane_outcomes = (
            data_plane_packet_propagator.get_as_outcomes_for_data_plane_packet(
                dest_ip_addr=scenario.dest_ip_addr,
                simulation_engine=engine,
                legitimate_origin_asns=scenario.legitimate_origin_asns,
                attacker_asns=scenario.attacker_asns,
                scenario=scenario,
            )
        )
        # NOTE: We used to include the data_tracker, but we don't now for three reasons:
        # 1. Any time the format of the metrics change, the tests break, breaking
        #    backwards compatibility
        # 2. Nobody looked at them. Like, ever. I.e. they were useless to test
        # 3. Many simulators track a very different set of metrics with no clear format
        self._store_data(engine=engine, asn_to_packet_outcome_dict=data_plane_outcomes)
        self._generate_diagrams(scenario, dpi=dpi)
        self._compare_against_ground_truth()

    def _get_engine_and_scenario(self):
        """Gets the engine and scenario"""
        engine = self._get_engine()
        scenario = self._get_scenario(engine=engine)
        scenario.setup_engine(engine)
        return engine, scenario

    def _get_engine(self):
        """Gets the engine"""
        return SimulationEngine(as_graph=self.conf.as_graph)

    def _get_scenario(self, engine: SimulationEngine):
        """Gets the scenario"""
        route_validator = next(
            iter(engine.as_graph.as_dict.values())
        ).policy.route_validator
        return self.conf.scenario_config.ScenarioCls(
            scenario_config=self.conf.scenario_config,
            engine=engine,
            route_validator=route_validator,
        )

    def _store_data(
        self, engine: SimulationEngine, asn_to_packet_outcome_dict: dict[int, Outcomes]
    ):
        """Stores the engine and outcomes.

        Always stores the guess, and optionally overwrites ground truth.
        """
        self.engine_guess_path.write_text(json.dumps(engine.to_json()))
        self.outcomes_guess_path.write_text(json.dumps(asn_to_packet_outcome_dict))
        # Only write the ground truth if we're comparing against it
        if self.compare_against_ground_truth and (
            self.overwrite or not self.engine_gt_path.exists()
        ):
            self.engine_gt_path.write_text(json.dumps(engine.to_json()))
        if self.compare_against_ground_truth and (
            self.overwrite or not self.outcomes_gt_path.exists()
        ):
            self.outcomes_gt_path.write_text(json.dumps(asn_to_packet_outcome_dict))

    def _generate_diagrams(self, scenario: Scenario, dpi: int | None = None):
        """Generates the diagrams"""

        if not self.write_diagrams:
            return

        vals = [
            (
                self.engine_guess_path,
                self.outcomes_guess_path,
                self.diagram_guess_path,
                "",
            ),
        ]
        # Only write the ground truth if we're comparing against it
        if self.compare_against_ground_truth:
            vals.append(
                (
                    self.engine_gt_path,
                    self.outcomes_gt_path,
                    self.diagram_gt_path,
                    " (ground truth) ",
                )
            )

        for engine_path, packet_outcomes_path, diagram_path, name in vals:
            Diagram().run(
                engine=SimulationEngine.from_json(json.loads(engine_path.read_text())),
                scenario=scenario,
                packet_outcomes={
                    int(asn): Outcomes(outcome)
                    for asn, outcome in json.loads(
                        packet_outcomes_path.read_text()
                    ).items()
                },
                name=self.conf.name + name,
                description=self.conf.diagram_desc,
                diagram_ranks=self.conf.diagram_ranks,
                path=diagram_path,
                dpi=dpi,
            )

    def _compare_against_ground_truth(self):
        if not self.compare_against_ground_truth:
            return

        """Compares the guesses against ground truth for engine and packet outcomes"""
        engine_guess = SimulationEngine.from_json(
            json.loads(self.engine_guess_path.read_text())
        )
        engine_gt = SimulationEngine.from_json(
            json.loads(self.engine_gt_path.read_text())
        )
        assert engine_guess == engine_gt, (
            "Engine guess does not match engine ground truth"
        )

        outcomes_guess = json.loads(self.outcomes_guess_path.read_text())
        outcomes_gt = json.loads(self.outcomes_gt_path.read_text())
        assert outcomes_guess == outcomes_gt, (
            "Outcomes guess does not match outcomes ground truth"
        )

    ###################
    # Path Properties #
    ###################

    @property
    def engine_guess_path(self) -> Path:
        return self.storage_dir / "engine_guess.json"

    @property
    def engine_gt_path(self) -> Path:
        return self.storage_dir / "engine_gt.json"

    @property
    def outcomes_guess_path(self) -> Path:
        return self.storage_dir / "outcomes_guess.json"

    @property
    def outcomes_gt_path(self) -> Path:
        return self.storage_dir / "outcomes_gt.json"

    @property
    def diagram_guess_path(self) -> Path:
        return self.storage_dir / "diagram_guess.gv"

    @property
    def diagram_gt_path(self) -> Path:
        return self.storage_dir / "diagram_gt.gv"



from collections import defaultdict, deque
from pathlib import Path
from typing import TYPE_CHECKING

from graphviz import Digraph

from bgpsimulator.shared import Outcomes, Settings
from bgpsimulator.simulation_engine import SimulationEngine
from bgpsimulator.simulation_framework import Scenario

if TYPE_CHECKING:
    from bgpsimulator.as_graphs import AS


class Diagram:
    """Generates a diagram of the engine run"""

    def __init__(self) -> None:
        self.dot: Digraph = Digraph(format="png")

    def run(
        self,
        engine: SimulationEngine,
        scenario: Scenario,
        packet_outcomes: dict[int, Outcomes],
        name: str,
        description: str,
        diagram_ranks: list[list[int]],
        path: Path,
        view: bool = False,
        dpi: int | None = None,
    ) -> None:
        """Runs the diagram"""
        self._add_legend(packet_outcomes, scenario)
        display_full_prefix_bool = self._get_display_full_prefix_bool(scenario)
        self._add_ases(engine, packet_outcomes, scenario, display_full_prefix_bool)
        self._add_edges(engine)
        diagram_ranks = diagram_ranks or self._get_default_diagram_ranks(engine)
        self._add_diagram_ranks(diagram_ranks, engine)
        self._add_description(name, description)
        self._render(path=path, view=view, dpi=dpi)

    def _add_legend(
        self, packet_outcomes: dict[int, Outcomes], scenario: Scenario
    ) -> None:
        """Adds legend to the graph with outcome counts"""

        attacker_success_count = sum(
            1 for x in packet_outcomes.values() if x == Outcomes.ATTACKER_SUCCESS.value
        )
        victim_success_count = sum(
            1
            for x in packet_outcomes.values()
            if x == Outcomes.LEGITIMATE_ORIGIN_SUCCESS.value
        )
        disconnect_count = sum(
            1 for x in packet_outcomes.values() if x == Outcomes.DISCONNECTED.value
        )
        looping_count = sum(
            1 for x in packet_outcomes.values() if x == Outcomes.DATA_PLANE_LOOP.value
        )
        html = f"""<
              <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0" CELLPADDING="4">
              <TR>
          <TD COLSPAN="2" BORDER="0">(For Destination of {scenario.dest_ip_addr})</TD>
              </TR>
              <TR>
          <TD BGCOLOR="#ff6060:white">&#128520; ATTACKER SUCCESS &#128520;</TD>
                <TD>{attacker_success_count}</TD>
              </TR>
              <TR>
         <TD BGCOLOR="#90ee90:white">&#128519; LEGITIMATE ORIGIN SUCCESS &#128519;</TD>
                <TD>{victim_success_count}</TD>
              </TR>
        """
        if disconnect_count:
            html += f"""
            <TR>
            <TD BGCOLOR="grey:white">&#10041; DISCONNECTED &#10041;</TD>
                    <TD>{disconnect_count}</TD>
                  </TR>
            """
        if looping_count:
            html += f"""
            <TR>
            <TD BGCOLOR="yellow:white">&#8734; LOOPING &#8734;</TD>
                    <TD>{disconnect_count}</TD>
                  </TR>
            """

        # ROAs takes up the least space right underneath the legend
        # which is why we have this here instead of a separate node
        html += """
              <TR>
                <TD COLSPAN="2" BORDER="0">ROAs (prefix, origin, max_len)</TD>
              </TR>
              """
        for roa in scenario.roas:
            html += f"""
              <TR>
                <TD>{roa.prefix}</TD>
                <TD>{roa.origin}</TD>
                <TD>{roa.max_length}</TD>
              </TR>"""
        html += """</TABLE>>"""

        self.dot.node(
            "Legend",
            html,
            shape="plaintext",
            color="black",
            style="filled",
            fillcolor="white",
        )

    def _get_display_full_prefix_bool(self, scenario: Scenario) -> bool:
        """If there are multiple prefixes for the same prefixlen, display full prefix

        else just display the prefix len as an abbreviation
        """

        prefix_len_to_addr = defaultdict(set)
        for asn, anns_list in scenario.seed_asn_ann_dict.items():
            for ann in anns_list:
                prefix_len_to_addr[ann.prefix.prefixlen].add(ann.prefix.network_address)
        return any(len(v) > 1 for v in prefix_len_to_addr.values())

    def _add_ases(
        self,
        engine: SimulationEngine,
        packet_outcomes: dict[int, Outcomes],
        scenario: Scenario,
        display_full_prefix_bool: bool,
    ) -> None:
        """Adds the ASes to the graph"""

        # First add all nodes to the graph
        for as_obj in engine.as_graph:
            self._encode_as_obj_as_node(
                as_obj, engine, packet_outcomes, scenario, display_full_prefix_bool
            )

    def _encode_as_obj_as_node(
        self,
        as_obj: "AS",
        engine: SimulationEngine,
        packet_outcomes: dict[int, Outcomes],
        scenario: Scenario,
        display_full_prefix_bool: bool,
    ) -> None:
        kwargs = dict()

        html = self._get_html(
            as_obj, engine, packet_outcomes, scenario, display_full_prefix_bool
        )

        kwargs = self._get_kwargs(as_obj, engine, packet_outcomes, scenario)

        self.dot.node(str(as_obj.asn), html, **kwargs)

    def _get_html(
        self,
        as_obj: "AS",
        engine: SimulationEngine,
        packet_outcomes: dict[int, Outcomes],
        scenario: Scenario,
        display_full_prefix_bool: bool,
    ) -> str:
        colspan = 3
        asn_str = str(as_obj.asn)
        if as_obj.asn in scenario.legitimate_origin_asns:
            asn_str = "&#128519;" + asn_str + "&#128519;"
        elif as_obj.asn in scenario.attacker_asns:
            asn_str = "&#128520;" + asn_str + "&#128520;"

        used_settings = [
            setting
            for setting, value in zip(Settings, as_obj.policy.settings, strict=False)
            if value and setting != Settings.BGP_FULL
        ]

        html = f"""<
            <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0" CELLPADDING="{colspan}">
            <TR>
            <TD COLSPAN="{colspan}" BORDER="0">{asn_str}</TD>
            </TR>
            """
        for setting in used_settings:
            html += f"""
            <TR>
            <TD COLSPAN="{colspan}" BORDER="0">({setting.name})</TD>
            </TR>"""
        local_rib_anns = tuple(as_obj.policy.local_rib.values())
        local_rib_anns = tuple(
            sorted(
                local_rib_anns,
                key=lambda x: x.prefix.num_addresses,
                reverse=True,
            )
        )
        if len(local_rib_anns) > 0:
            html += f"""<TR>
                        <TD COLSPAN="{colspan}">Local RIB</TD>
                      </TR>"""

            for ann in local_rib_anns:
                if display_full_prefix_bool:
                    prefix_display = str(ann.prefix)
                else:
                    prefix_display = "/" + str(ann.prefix).split("/")[-1]
                path = "-".join(str(x) for x in ann.as_path)
                html += f"""<TR>
                            <TD COLSPAN="1">{prefix_display}</TD>
                            <TD COLSPAN="2">{path}</TD>
                            """
                html += """</TR>"""
        html += "</TABLE>>"
        return html

    def _get_kwargs(
        self,
        as_obj: "AS",
        engine: SimulationEngine,
        packet_outcomes: dict[int, Outcomes],
        scenario: Scenario,
    ) -> dict[str, str]:
        kwargs = {
            "color": "black",
            "style": "filled",
            "fillcolor": "white",
            "gradientangle": "270",
        }

        # If the as obj is the attacker
        if as_obj.asn in scenario.attacker_asns:
            kwargs.update({"fillcolor": "#ff6060", "shape": "doublecircle"})
            if any(
                v
                for k, v in zip(Settings, as_obj.policy.settings, strict=False)
                if v and k != Settings.BGP_FULL
            ):
                kwargs["shape"] = "doubleoctagon"
            # If people complain about the red being too dark lol:
            kwargs.update({"fillcolor": "#FF7F7F"})
            # kwargs.update({"fillcolor": "#ff4d4d"})
        # As obj is the victim
        elif as_obj.asn in scenario.legitimate_origin_asns:
            kwargs.update({"fillcolor": "#90ee90", "shape": "doublecircle"})
            if any(
                v
                for k, v in zip(Settings, as_obj.policy.settings, strict=False)
                if v and k != Settings.BGP_FULL
            ):
                kwargs["shape"] = "doubleoctagon"

        # As obj is not attacker or victim
        else:
            if packet_outcomes[as_obj.asn] == Outcomes.ATTACKER_SUCCESS.value:
                kwargs.update({"fillcolor": "#ff6060:yellow"})
            elif (
                packet_outcomes[as_obj.asn] == Outcomes.LEGITIMATE_ORIGIN_SUCCESS.value
            ):
                kwargs.update({"fillcolor": "#90ee90:white"})
            elif packet_outcomes[as_obj.asn] == Outcomes.DISCONNECTED.value:
                kwargs.update({"fillcolor": "grey:white"})
            elif packet_outcomes[as_obj.asn] == Outcomes.DATA_PLANE_LOOP.value:
                kwargs.update({"fillcolor": "yellow:white"})

            if any(
                v
                for k, v in zip(Settings, as_obj.policy.settings, strict=False)
                if v and k != Settings.BGP_FULL
            ):
                kwargs["shape"] = "octagon"
        return kwargs

    def _add_edges(self, engine: SimulationEngine):
        # Then add all connections to the graph
        # Starting with provider to customer
        for as_obj in engine.as_graph:
            # Add provider customer edges
            for customer_obj in as_obj.customers:
                self.dot.edge(str(as_obj.asn), str(customer_obj.asn))
            # Add peer edges
            # Only add if the largest asn is the curren as_obj to avoid dups
            for peer_obj in as_obj.peers:
                if as_obj.asn > peer_obj.asn:
                    self.dot.edge(
                        str(as_obj.asn),
                        str(peer_obj.asn),
                        dir="none",
                        style="dashed",
                        penwidth="2",
                    )

    def _get_default_diagram_ranks(self, engine) -> list[list[int]]:
        """
        Return ranks (list of ASN lists) for drawing the AS graph.

        * Peered ASNs share a rank.
        * A provider is always strictly above each customer.
        * Multihomed customers sit just below their highest provider.
        * Output ordering is deterministic.
        """
        # --------------------------------------------------------------
        # 1.  Gather raw edges (ASNs only, no AS objects)
        # --------------------------------------------------------------
        as_dict = engine.as_graph.as_dict  # dict[int, ASNode]
        peer_edges: set[tuple[int, int]] = set()
        provider_edges: set[tuple[int, int]] = set()  # (provider, customer)

        for asn, node in as_dict.items():
            # Peers (undirected) – store ints (node.peers is a set of AS-objects)
            for peer_obj in node.peers:
                peer_asn = peer_obj.asn
                if peer_asn in as_dict:  # ignore external ASNs
                    peer_edges.add(tuple(sorted((asn, peer_asn))))

            # Providers (directed) – provider ➜ customer (= asn)
            for prov_obj in node.providers:
                prov_asn = prov_obj.asn
                if prov_asn in as_dict:
                    provider_edges.add((prov_asn, asn))

        # --------------------------------------------------------------
        # 2.  Collapse peer clusters with Union–Find
        # --------------------------------------------------------------
        parent: dict[int, int] = {}

        def find(x: int) -> int:
            parent.setdefault(x, x)
            if parent[x] != x:
                parent[x] = find(parent[x])
            return parent[x]

        def union(a: int, b: int) -> None:
            pa, pb = find(a), find(b)
            if pa != pb:
                parent[pb] = pa

        for u, v in peer_edges:
            union(u, v)

        cluster_of = {asn: find(asn) for asn in as_dict}

        # --------------------------------------------------------------
        # 3.  Build provider-customer DAG between clusters
        # --------------------------------------------------------------
        children: dict[int, set[int]] = defaultdict(set)
        indeg: dict[int, int] = defaultdict(int)

        for prov_asn, cust_asn in provider_edges:
            cp, cc = cluster_of[prov_asn], cluster_of[cust_asn]
            if cp == cc:  # provider inside its own peer-cluster → ignore
                continue
            if cc not in children[cp]:
                children[cp].add(cc)
                indeg[cc] += 1

        all_clusters = set(cluster_of.values())
        rank: dict[int, int] = defaultdict(int)

        # --------------------------------------------------------------
        # 4.  Longest-path layering
        # --------------------------------------------------------------
        queue = deque(cl for cl in all_clusters if indeg[cl] == 0)

        while queue:
            cl = queue.popleft()
            for child in children.get(cl, ()):
                rank[child] = max(rank[child], rank[cl] + 1)
                indeg[child] -= 1
                if indeg[child] == 0:
                    queue.append(child)

        # Any cluster still showing indegree > 0 is in a cycle → push to top
        for cl, d in indeg.items():
            if d:  # d > 0
                rank[cl] = 0

        # --------------------------------------------------------------
        # 5.  Expand clusters back to ASNs and sort
        # --------------------------------------------------------------
        levels: dict[int, list[int]] = defaultdict(list)
        for asn, cl in cluster_of.items():
            levels[rank[cl]].append(asn)

        max_rank = max(levels) if levels else -1
        return [sorted(levels[r]) for r in range(max_rank + 1)]

    def _add_diagram_ranks(
        self, diagram_ranks: list[list[int]], engine: SimulationEngine
    ) -> None:
        # TODO: Refactor
        if diagram_ranks:
            for diagram_rank in diagram_ranks:
                with self.dot.subgraph() as s:
                    s.attr(rank="same")  # set all nodes to the same rank
                    previous_asn: str | None = None
                    for asn in diagram_rank:
                        assert isinstance(asn, int)
                        s.node(str(asn))
                        if previous_asn is not None:
                            # Add invisible edge to maintain static order
                            s.edge(str(previous_asn), str(asn), style="invis")
                        previous_asn = str(asn)
        else:
            for i, rank in enumerate(engine.as_graph.propagation_ranks):
                g = Digraph(f"Propagation_rank_{i}")
                g.attr(rank="same")
                for as_obj in rank:
                    g.node(str(as_obj.asn))
                self.dot.subgraph(g)

    def _add_description(self, name: str, description: str) -> None:
        if name or description:
            # https://stackoverflow.com/a/57461245/8903959
            self.dot.attr(label=f"{name}\n{description}")

    def _render(
        self, path: Path | None = None, view: bool = False, dpi: int | None = None
    ) -> None:
        if dpi:
            self.dot.attr(dpi=str(dpi))
        self.dot.render(path, view=view)



from typing import TYPE_CHECKING

from bgpsimulator.shared.enums import Relationships

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class EnforceFirstAS:
    """A Policy that enforces the first AS in the AS-Path to be the origin AS"""

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Ensures the first ASN in the AS-Path is a neighbor

        NOTE: normally this would check for an exact match, but since we don't
        store which ASN the announcement came from anywhere, we just check if it
        is a neighbor to simulate, since we've always picked attackers at the edge
        """

        return (
            ann.next_hop_asn == ann.as_path[0]
            # Super janky, TODO
            and ann.next_hop_asn in policy.as_.neighbor_asns
        )



import gc
from pathlib import Path
from statistics import mean
from typing import Any

import matplotlib as mpl
import matplotlib.pyplot as plt

from bgpsimulator.simulation_framework.data_tracker import LineFilter

from .line import Line


class LineChart:
    """A line chart"""

    def __init__(
        self,
        line_filter: LineFilter,
        lines: list[Line],
        title: str,
        xlabel: str,
        ylabel: str,
        xlim: tuple[float, float],
        ylim: tuple[float, float],
        legend_loc: str = "best",
    ) -> None:
        self.line_filter = line_filter
        self.lines = lines
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.xlim = xlim
        self.ylim = ylim
        self.legend_loc = legend_loc

    def write_graph(self, path: Path) -> None:
        """Writes the graph to a file"""

        mpl.use("Agg")
        fig, ax = plt.subplots()
        fig.set_dpi(300)
        plt.rcParams.update({"font.size": 14, "lines.markersize": 10})

        # Set labels
        # ax.set_ylabel(self.ylabel)
        # ax.set_xlabel(self.xlabel)
        for line in self.lines:
            ax.errorbar(
                line.xs,
                line.ys,
                line.yerrs,
                label=line.label,
                marker=line.marker,
                ls=line.ls,
                color=line.color,
            )

        # Set X and Y axis size
        ax.set_xlim(*self.xlim)
        ax.set_ylim(*self.ylim)

        handles, labels = ax.get_legend_handles_labels()
        label_to_handle_dict = dict(zip(labels, handles, strict=False))

        mean_y_dict = {line.label: mean(line.ys) for line in self.lines}
        sorted_labels = sorted(
            label_to_handle_dict,
            key=lambda label: mean_y_dict[label],
            reverse=True,
        )
        sorted_handles = [label_to_handle_dict[lbl] for lbl in sorted_labels]

        ax.legend(sorted_handles, sorted_labels)
        plt.tight_layout()
        plt.savefig(path)

        # https://stackoverflow.com/a/33343289/8903959
        ax.cla()
        plt.cla()
        plt.clf()
        # If you just close the fig, on machines with many CPUs and trials,
        # there is some sort of a memory leak that occurs. See stackoverflow
        # comment above
        plt.close(fig)
        # If you are running one simulation after the other, matplotlib
        # basically leaks memory. I couldn't find the original issue, but
        # here is a note in one of their releases saying to just call the garbage
        # collector: https://matplotlib.org/stable/users/prev_whats_new/
        # whats_new_3.6.0.html#garbage-collection-is-no-longer-run-on-figure-close
        # and here is the stackoverflow post on this topic:
        # https://stackoverflow.com/a/33343289/8903959
        # Even if this works without garbage collection in 3.5.2, that will break
        # as soon as we upgrade to the latest matplotlib which no longer does
        # If you run the simulations on a machine with many cores and lots of trials,
        # this bug leaks enough memory to crash the server, so we must garbage collect
        gc.collect()

    def to_json(self) -> dict[str, Any]:
        return {
            "line_filter": self.line_filter.to_json(),
            "title": self.title,
            "xlabel": self.xlabel,
            "ylabel": self.ylabel,
            "xlim": self.xlim,
            "ylim": self.ylim,
            "legend_loc": self.legend_loc,
            "lines": [line.to_json() for line in self.lines],
        }

    @classmethod
    def from_json(cls, json_data: dict[str, Any]) -> "LineChart":
        return cls(
            line_filter=LineFilter.from_json(json_data["line_filter"]),
            title=json_data["title"],
            xlabel=json_data["xlabel"],
            ylabel=json_data["ylabel"],
            xlim=json_data["xlim"],
            ylim=json_data["ylim"],
            legend_loc=json_data["legend_loc"],
            lines=[Line.from_json(line) for line in json_data["lines"]],
        )



from typing import TYPE_CHECKING

from bgpsimulator.shared import PolicyPropagateInfo, Relationships

if TYPE_CHECKING:
    from bgpsimulator.as_graphs import AS
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class OnlyToCustomers:
    """A Policy that deploys only to customers, RFC 9234"""

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Returns validity for OTC attributes (RFC 9234)"""

        if (
            ann.only_to_customers
            and from_rel == Relationships.PEERS
            and ann.next_hop_asn != ann.only_to_customers
        ) or (ann.only_to_customers and from_rel == Relationships.CUSTOMERS):
            return False
        else:
            return True

    @staticmethod
    def get_policy_propagate_vals(
        policy: "Policy",
        neighbor_as_obj: "AS",
        ann: "Ann",
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> PolicyPropagateInfo:
        """If propagating to custmoers and only_to_customers isn't set, set it"""

        if propagate_to in (Relationships.CUSTOMERS, Relationships.PEERS):
            ann = ann.copy(only_to_customers=policy.as_.asn)
            policy.process_outgoing_ann(neighbor_as_obj, ann, propagate_to, send_rels)
            return PolicyPropagateInfo(
                policy_propagate_bool=True, ann=ann, send_ann_bool=True
            )
        else:
            return PolicyPropagateInfo(
                policy_propagate_bool=False, ann=ann, send_ann_bool=False
            )



from bgpsimulator.as_graphs import AS
from bgpsimulator.shared import IPAddr, Outcomes, Relationships
from bgpsimulator.simulation_engine import Announcement as Ann
from bgpsimulator.simulation_engine import SimulationEngine

from .scenarios import Scenario


class DataPlanePacketPropagator:
    def get_as_outcomes_for_data_plane_packet(
        self,
        dest_ip_addr: IPAddr,
        simulation_engine: SimulationEngine,
        legitimate_origin_asns: set[int],
        attacker_asns: set[int],
        scenario: Scenario,
    ) -> dict[int, Outcomes]:
        """For each AS, determine the outcome of a packet sent to the
        destination IP address

        See ROV++ paper as to why the control plane is a bad metric;
        so we only focus on the data plane
        """

        outcomes: dict[int, Outcomes] = dict()
        for as_obj in simulation_engine.as_graph:
            self.store_as_data_plane_outcomes(
                as_obj,
                simulation_engine,
                dest_ip_addr,
                outcomes=outcomes,
                visited_asns=set(),
                legitimate_origin_asns=legitimate_origin_asns,
                attacker_asns=attacker_asns,
                scenario=scenario,
            )
        return outcomes

    def store_as_data_plane_outcomes(
        self,
        as_obj: AS,
        engine: SimulationEngine,
        dest_ip_addr: IPAddr,
        outcomes: dict[int, Outcomes],
        visited_asns: set[int],
        legitimate_origin_asns: set[int],
        attacker_asns: set[int],
        scenario: Scenario,
    ):
        """Recursively stores the outcomes of the AS on the data plane"""

        if as_obj.asn in outcomes:
            return outcomes[as_obj.asn]
        else:
            most_specific_ann = as_obj.policy.get_most_specific_ann(dest_ip_addr)
            outcome = self._determine_data_plane_outcome(
                as_obj,
                engine,
                dest_ip_addr,
                most_specific_ann,
                visited_asns,
                legitimate_origin_asns,
                attacker_asns,
                scenario,
            )
            if outcome == Outcomes.UNDETERMINED:
                # outcome won't ever be undetermined if most_specific_ann is None
                assert most_specific_ann, "for mypy"

                # next as in the AS path to traceback to
                # Ignore type because only way for this to be here
                # Is if the most specific "Ann" was NOT None.
                next_as = engine.as_graph.as_dict[
                    # NOTE: this is the next hop,
                    # not the next ASN in the AS PATH
                    # This is more in line with real BGP and allows for more
                    # advanced types of hijacks such as origin spoofing hijacks
                    most_specific_ann.next_hop_asn
                ]
                visited_asns.add(as_obj.asn)
                outcome = self.store_as_data_plane_outcomes(
                    next_as,
                    engine,
                    dest_ip_addr,
                    outcomes,
                    visited_asns,
                    legitimate_origin_asns,
                    attacker_asns,
                    scenario,
                )
            outcomes[as_obj.asn] = outcome
            return outcome

    def _determine_data_plane_outcome(
        self,
        as_obj: AS,
        engine: SimulationEngine,
        dest_ip_addr: IPAddr,
        most_specific_ann: Ann | None,
        visited_asns: set[int],
        legitimate_origin_asns: set[int],
        attacker_asns: set[int],
        scenario: Scenario,
    ) -> Outcomes:
        """Determines the outcome at an AS"""

        if as_obj.asn in attacker_asns:
            return Outcomes.ATTACKER_SUCCESS
        elif as_obj.asn in legitimate_origin_asns:
            return Outcomes.LEGITIMATE_ORIGIN_SUCCESS
        # End of traceback
        elif (
            most_specific_ann is None
            or len(most_specific_ann.as_path) == 1
            or most_specific_ann.recv_relationship.value == Relationships.ORIGIN.value
            or most_specific_ann.next_hop_asn == as_obj.asn
            or not as_obj.policy.passes_sav(dest_ip_addr, most_specific_ann)
        ):
            return Outcomes.DISCONNECTED
        elif (as_obj.asn in visited_asns) or (len(visited_asns) > 64):
            return Outcomes.DATA_PLANE_LOOP
        else:
            return Outcomes.UNDETERMINED



from typing import TYPE_CHECKING

from bgpsimulator.shared import PolicyPropagateInfo, Relationships, Settings

from .bgpsec import BGPSec

if TYPE_CHECKING:
    from bgpsimulator.as_graphs import AS
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class BGPiSecTransitive:
    """A Policy that deploys BGPiSec-Transitive as defined in the BGPiSec paper

    Extends BGPSec, aside from the path preference mechanism, which BGPiSec reverts
    to BGP. Additionally, this is only the transitive signatures of BGPiSec, not the
    entire policy
    """

    @staticmethod
    def get_modified_seed_ann(policy: "Policy", ann: "Ann") -> "Ann":
        return BGPSec.get_modified_seed_ann(policy, ann)

    @staticmethod
    def get_policy_propagate_vals(
        policy: "Policy",
        neighbor_as: "AS",
        ann: "Ann",
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> PolicyPropagateInfo:
        """Always set the BGPSec vals when propagating for BGPiSec"""

        send_ann = ann.copy(
            bgpsec_next_asn=neighbor_as.asn, bgpsec_as_path=ann.bgpsec_as_path
        )
        return PolicyPropagateInfo(
            policy_propagate_bool=True, ann=send_ann, send_ann_bool=True
        )

    @staticmethod
    def process_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> "Ann":
        """Sets the bgpsec_as_path always for transitive signatures"""

        return ann.copy(bgpsec_as_path=(policy.as_.asn, *ann.bgpsec_as_path))

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Determines bgp-isec transitive validity

        If any ASes along the AS path are adopting and are not in the bgpsec_as_path,
        that means those ASes didn't add signatures, therefore the ann is missing
        signatures and should be dropped
        """

        as_graph = policy.as_.as_graph
        bgpsec_signatures = ann.bgpsec_as_path
        for asn in ann.as_path:
            if asn not in bgpsec_signatures and (
                as_graph.as_dict[asn].policy.settings[Settings.BGP_I_SEC]
                or as_graph.as_dict[asn].policy.settings[Settings.BGP_I_SEC_TRANSITIVE]
            ):
                return False
        return True



from typing import TYPE_CHECKING

from bgpsimulator.shared import PolicyPropagateInfo, Relationships

if TYPE_CHECKING:
    from bgpsimulator.as_graphs import AS
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class FirstASNStrippingPrefixHijackCustomers:
    """A Policy that performs prefix origin hijacks against customers.

    But strips the first ASN.

    ASPA and enforce-first-asn prevent this
    """

    @staticmethod
    def get_policy_propagate_vals(
        policy: "Policy",
        neighbor_as: "AS",
        ann: "Ann",
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> PolicyPropagateInfo:
        """Performs prefix origin hijacks against customers, but strips the first ASN"""

        # This ann is originating from here, the attacker, so it's an attacker's ann
        # If as path length is 1 (like it would be against BGP), don't modify it
        if (
            propagate_to == Relationships.CUSTOMERS
            and ann.recv_relationship == Relationships.ORIGIN
            and len(ann.as_path) > 1
        ):
            # Only need when sending to customers
            return PolicyPropagateInfo(
                policy_propagate_bool=True,
                ann=ann.copy(as_path=(ann.origin,)),
                send_ann_bool=True,
            )
        else:
            return PolicyPropagateInfo(
                policy_propagate_bool=False,
                ann=ann,
                send_ann_bool=True,
            )



from bgpsimulator.as_graphs.as_graph import ASGraph
from bgpsimulator.engine_runner import EngineRunConfig
from bgpsimulator.shared import CommonASNs, Settings
from bgpsimulator.simulation_framework import ScenarioConfig, SuperprefixPrefixHijack

graph_data = {
    "ases": {
        str(CommonASNs.LEGITIMATE_ORIGIN): {
            "asn": CommonASNs.LEGITIMATE_ORIGIN,
            "customer_asns": [],
            "peer_asns": [],
            "provider_asns": [2, 4, 10],
        },
        str(CommonASNs.ATTACKER): {
            "asn": CommonASNs.ATTACKER,
            "customer_asns": [],
            "peer_asns": [],
            "provider_asns": [1, 2],
        },
        "1": {
            "asn": 1,
            "customer_asns": [CommonASNs.ATTACKER],
            "peer_asns": [],
            "provider_asns": [5, 8],
        },
        "2": {
            "asn": 2,
            "customer_asns": [CommonASNs.ATTACKER, CommonASNs.LEGITIMATE_ORIGIN],
            "peer_asns": [],
            "provider_asns": [8],
        },
        "3": {
            "asn": 3,
            "customer_asns": [],
            "peer_asns": [9],
            "provider_asns": [],
        },
        "4": {
            "asn": 4,
            "customer_asns": [CommonASNs.LEGITIMATE_ORIGIN],
            "peer_asns": [],
            "provider_asns": [9],
        },
        "5": {
            "asn": 5,
            "customer_asns": [1],
            "peer_asns": [],
            "provider_asns": [],
        },
        "8": {
            "asn": 8,
            "customer_asns": [1, 2],
            "peer_asns": [9],
            "provider_asns": [11],
        },
        "9": {
            "asn": 9,
            "customer_asns": [4],
            "peer_asns": [8, 10, 3],
            "provider_asns": [11],
        },
        "10": {
            "asn": 10,
            "customer_asns": [CommonASNs.LEGITIMATE_ORIGIN],
            "peer_asns": [9],
            "provider_asns": [11, 12],
        },
        "11": {
            "asn": 11,
            "customer_asns": [8, 9, 10],
            "peer_asns": [],
            "provider_asns": [],
        },
        "12": {
            "asn": 12,
            "customer_asns": [10],
            "peer_asns": [],
            "provider_asns": [],
        },
    },
}

# Create the engine run config
ex_config_011 = EngineRunConfig(
    name="ex_011_superprefix_prefix_hijack_rov_simple",
    scenario_config=ScenarioConfig(
        label="rov",
        ScenarioCls=SuperprefixPrefixHijack,
        override_attacker_asns={CommonASNs.ATTACKER},
        override_legitimate_origin_asns={CommonASNs.LEGITIMATE_ORIGIN},
        # AS 9 uses ROV
        override_base_settings={9: {Settings.ROV: True}},
    ),
    as_graph=ASGraph(graph_data),
    diagram_desc="Superprefix + Prefix Hijack with ROV simple",
)



from bgpsimulator.as_graphs.as_graph import ASGraph
from bgpsimulator.engine_runner import EngineRunConfig
from bgpsimulator.shared import CommonASNs, Settings
from bgpsimulator.simulation_framework import AccidentalRouteLeak, ScenarioConfig

# Graph to test OTC from a peer
# 777 - 666 - 1
graph_data = {
    "ases": {
        str(CommonASNs.LEGITIMATE_ORIGIN): {
            "asn": CommonASNs.LEGITIMATE_ORIGIN,
            "customer_asns": [],
            "peer_asns": [CommonASNs.ATTACKER],
            "provider_asns": [],
        },
        str(CommonASNs.ATTACKER): {
            "asn": CommonASNs.ATTACKER,
            "customer_asns": [],
            "peer_asns": [CommonASNs.LEGITIMATE_ORIGIN, 1],
            "provider_asns": [],
        },
        "1": {
            "asn": 1,
            "customer_asns": [],
            "peer_asns": [CommonASNs.ATTACKER],
            "provider_asns": [],
        },
    },
}

# Create the engine run config
ex_config_031 = EngineRunConfig(
    name="ex_031_route_leak_peer_with_otc_simple",
    scenario_config=ScenarioConfig(
        label="otc",
        ScenarioCls=AccidentalRouteLeak,
        override_attacker_asns={CommonASNs.ATTACKER},
        override_legitimate_origin_asns={CommonASNs.LEGITIMATE_ORIGIN},
        # AS 1 and VICTIM use OnlyToCustomers
        override_base_settings={
            1: {Settings.ONLY_TO_CUSTOMERS: True},
            CommonASNs.LEGITIMATE_ORIGIN: {Settings.ONLY_TO_CUSTOMERS: True},
        },
    ),
    as_graph=ASGraph(graph_data),
    diagram_desc="Accidental route leak to a peer with OTC Simple",
)



from bgpsimulator.as_graphs.as_graph import ASGraph
from bgpsimulator.engine_runner import EngineRunConfig
from bgpsimulator.shared import CommonASNs, Settings
from bgpsimulator.simulation_framework import ForgedOriginPrefixHijack, ScenarioConfig

# Custom graph for this test
graph_data = {
    "ases": {
        str(CommonASNs.LEGITIMATE_ORIGIN): {
            "asn": CommonASNs.LEGITIMATE_ORIGIN,
            "customer_asns": [],
            "peer_asns": [],
            "provider_asns": [3],
        },
        str(CommonASNs.ATTACKER): {
            "asn": CommonASNs.ATTACKER,
            "customer_asns": [],
            "peer_asns": [],
            "provider_asns": [1, 2],
        },
        "1": {
            "asn": 1,
            "customer_asns": [CommonASNs.ATTACKER],
            "peer_asns": [],
            "provider_asns": [5],
        },
        "2": {
            "asn": 2,
            "customer_asns": [CommonASNs.ATTACKER],
            "peer_asns": [],
            "provider_asns": [4],
        },
        "3": {
            "asn": 3,
            "customer_asns": [CommonASNs.LEGITIMATE_ORIGIN],
            "peer_asns": [],
            "provider_asns": [],
        },
        "4": {
            "asn": 4,
            "customer_asns": [2],
            "peer_asns": [],
            "provider_asns": [5],
        },
        "5": {
            "asn": 5,
            "customer_asns": [4, 1],
            "peer_asns": [],
            "provider_asns": [],
        },
    },
}

# Create the engine run config
ex_config_029 = EngineRunConfig(
    name="ex_029_aspa_weirdness",
    scenario_config=ScenarioConfig(
        label="aspa",
        ScenarioCls=ForgedOriginPrefixHijack,
        override_attacker_asns={CommonASNs.ATTACKER},
        override_legitimate_origin_asns={CommonASNs.LEGITIMATE_ORIGIN},
        # AS 4 and VICTIM use ASPA
        override_base_settings={
            4: {Settings.ASPA: True},
            CommonASNs.LEGITIMATE_ORIGIN: {Settings.ASPA: True},
        },
    ),
    as_graph=ASGraph(graph_data),
    diagram_desc="ASPA weirdness. rejects upstream, accepts downstream",
)



from typing import TYPE_CHECKING

from bgpsimulator.shared import PolicyPropagateInfo, Relationships

if TYPE_CHECKING:
    from bgpsimulator.as_graphs import AS
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class NeverPropagateWithdrawals:
    """A Policy that suppresses withdrawals"""

    @staticmethod
    def get_policy_propagate_vals(
        policy: "Policy",
        neighbor_as: "AS",
        ann: "Ann",
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> PolicyPropagateInfo:
        if ann.withdraw:
            return PolicyPropagateInfo(
                policy_propagate_bool=True, ann=ann, send_ann_bool=False
            )
        else:
            return PolicyPropagateInfo(
                policy_propagate_bool=False, ann=ann, send_ann_bool=True
            )



from typing import TYPE_CHECKING

from bgpsimulator.shared import Relationships, Settings

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class ProviderConeID:
    """A Policy that deploys Provider Cone ID as defined in the BGP-iSec paper

    For simplicity, we just put the full provider cone
    """

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Determines Provider Cone ID validity from customers"""

        if from_rel == Relationships.CUSTOMERS:
            as_dict = policy.as_.as_graph.as_dict
            provider_cone_asns = as_dict[ann.origin].provider_cone_asns
            # We don't look at the last ASN in the path, since that's the origin
            # The ASes ASN is also not yet in the announcement, so we add it here
            for asn in (policy.as_.asn, *ann.as_path[:-1]):
                # not in provider cone of the origin, and is adopting
                if asn not in provider_cone_asns and (
                    as_dict[asn].policy.settings[Settings.BGP_I_SEC]
                    or as_dict[asn].policy.settings[Settings.PROVIDER_CONE_ID]
                    or as_dict[asn].policy.settings[Settings.ASPAPP]
                ):
                    return False

        return True



from bgpsimulator.as_graphs.as_graph import ASGraph
from bgpsimulator.engine_runner import EngineRunConfig
from bgpsimulator.shared import CommonASNs, Settings
from bgpsimulator.simulation_framework import PrefixHijack, ScenarioConfig

# Graph to test OTC from a peer
# CommonASNs.LEGITIMATE_ORIGIN - CommonASNs.ATTACKER - 1
graph_data = {
    "ases": {
        "1": {"asn": 1, "customer_asns": [3, 5]},
        "2": {"asn": 2, "customer_asns": [4], "peer_asns": [5]},
        "3": {"asn": 3, "provider_asns": [1], "peer_asns": [4]},
        "4": {"asn": 4, "provider_asns": [2, 8], "peer_asns": [3]},
        "5": {
            "asn": 5,
            "provider_asns": [1],
            "customer_asns": [8, 9],
            "peer_asns": [2, 11],
        },
        "8": {"asn": 8, "provider_asns": [5], "customer_asns": [4], "peer_asns": [9]},
        "9": {
            "asn": 9,
            "provider_asns": [5],
            "customer_asns": [10, CommonASNs.ATTACKER, CommonASNs.LEGITIMATE_ORIGIN],
            "peer_asns": [8],
        },
        "10": {"asn": 10, "provider_asns": [9]},
        "11": {"asn": 11, "peer_asns": [5]},
        CommonASNs.ATTACKER: {"asn": CommonASNs.ATTACKER, "provider_asns": [9]},
        CommonASNs.LEGITIMATE_ORIGIN: {
            "asn": CommonASNs.LEGITIMATE_ORIGIN,
            "provider_asns": [9],
        },
    }
}

# Create the engine run config
ex_config_032 = EngineRunConfig(
    name="ex_032_graph_display_algo",
    scenario_config=ScenarioConfig(
        label="graph_display_algo",
        ScenarioCls=PrefixHijack,
        override_attacker_asns={CommonASNs.ATTACKER},
        override_legitimate_origin_asns={CommonASNs.LEGITIMATE_ORIGIN},
        # AS 1 and VICTIM use OnlyToCustomers
        override_base_settings={
            1: {Settings.ROV: True},
            CommonASNs.LEGITIMATE_ORIGIN: {Settings.ROV: True},
        },
    ),
    as_graph=ASGraph(graph_data),
    diagram_desc="Testing the graph display algorithm",
)



from .as_path_edge_filter import ASPathEdgeFilter
from .aspa import ASPA
from .aspawn import ASPAwN
from .asra import ASRA
from .bgp import BGP
from .only_to_customers import OnlyToCustomers
from .enforce_first_as import EnforceFirstAS
from .rov import ROV
from .path_end import PathEnd
from .peerlock_lite import PeerLockLite
from .rovpp_v1_lite import ROVPPV1Lite
from .rovpp_v2_lite import ROVPPV2Lite
from .rovpp_v2i_lite import ROVPPV2iLite
from .bgpsec import BGPSec
from .bgpisec_transitive import BGPiSecTransitive
from .provider_cone_id import ProviderConeID
from .origin_prefix_hijack_customers import OriginPrefixHijackCustomers
from .first_asn_stripping_prefix_hijack_customers import (
    FirstASNStrippingPrefixHijackCustomers,
)
from .peer_rov import PeerROV
from .aspapp import ASPAPP
from .rost import ROST, RoSTTrustedRepository
from .never_propagate_withdrawals import NeverPropagateWithdrawals
from .leaker import Leaker
from .announce_then_withdraw import AnnounceThenWithdraw

__all__ = [
    "ASPathEdgeFilter",
    "ASPA",
    "ASPAwN",
    "ASRA",
    "BGP",
    "OnlyToCustomers",
    "EnforceFirstAS",
    "ROV",
    "PathEnd",
    "PeerLockLite",
    "ROVPPV1Lite",
    "ROVPPV2Lite",
    "ROVPPV2iLite",
    "BGPSec",
    "BGPiSecTransitive",
    "ProviderConeID",
    "OriginPrefixHijackCustomers",
    "FirstASNStrippingPrefixHijackCustomers",
    "PeerROV",
    "ASPAPP",
    "ROST",
    "RoSTTrustedRepository",
    "NeverPropagateWithdrawals",
    "Leaker",
    "AnnounceThenWithdraw",
]



from collections import UserDict

from bgpsimulator.shared import Prefix
from bgpsimulator.simulation_engine import Announcement as Ann


class AdjRIBsOut(UserDict[int, dict[Prefix, Ann]]):
    """Incomming announcements for a BGP AS

    neighbor: {prefix: announcement}
    """

    def get_ann(self, neighbor_asn: int, prefix: Prefix) -> Ann | None:
        """Returns Ann for a given neighbor asn and prefix"""

        return self.data.get(neighbor_asn, dict()).get(prefix)

    def add_ann(self, neighbor_asn: int, ann: Ann) -> None:
        """Adds announcement to the ribs out"""

        if neighbor_asn in self.data:
            self.data[neighbor_asn][ann.prefix] = ann
        else:
            self.data[neighbor_asn] = {ann.prefix: ann}

    def remove_entry(self, neighbor_asn: int, prefix: Prefix) -> bool:
        """Removes ann from ribs out"""

        try:
            del self.data[neighbor_asn][prefix]
            return True
        except KeyError:
            return False

    def populated_neighbors(self) -> list[int]:
        """Return all neighbors from the ribs out"""

        return list(self.data.keys())

    def to_json(self) -> dict[int, dict[str, dict[str, dict[str, Ann]]]]:
        """Returns a JSON representation of the AdjRIBsOut"""

        return {
            neighbor_asn: {
                str(prefix): ann.to_json() for prefix, ann in prefix_anns.items()
            }
            for neighbor_asn, prefix_anns in self.data.items()
        }

    @classmethod
    def from_json(
        cls, json: dict[int, dict[str, dict[str, dict[str, Ann]]]]
    ) -> "AdjRIBsOut":
        """Returns a AdjRIBsOut from a JSON representation"""

        adj_ribs_out = cls()
        for neighbor_asn, prefix_anns in json.items():
            for _prefix, ann_json in prefix_anns.items():
                adj_ribs_out.add_ann(neighbor_asn, Ann.from_json(ann_json))
        return adj_ribs_out



from bgpsimulator.as_graphs.as_graph import ASGraph
from bgpsimulator.engine_runner import EngineRunConfig
from bgpsimulator.shared import CommonASNs, Settings
from bgpsimulator.simulation_framework import PassiveHijack, ScenarioConfig

# Graph to test OTC from a peer
# CommonASNs.LEGITIMATE_ORIGIN - CommonASNs.ATTACKER - 1
graph_data = {
    "ases": {
        "1": {
            "asn": 1,
            "customer_asns": [CommonASNs.ATTACKER],
            "provider_asns": [
                3,
            ],
        },
        "2": {
            "asn": 2,
            "customer_asns": [CommonASNs.LEGITIMATE_ORIGIN, CommonASNs.ATTACKER],
            "provider_asns": [3],
        },
        "3": {"asn": 3, "customer_asns": [1, 2]},
        CommonASNs.ATTACKER: {
            "asn": CommonASNs.ATTACKER,
            "provider_asns": [1, 2],
        },
        CommonASNs.LEGITIMATE_ORIGIN: {
            "asn": CommonASNs.LEGITIMATE_ORIGIN,
            "provider_asns": [2],
        },
    }
}

# Create the engine run config
ex_config_033 = EngineRunConfig(
    name="ex_033_leaker",
    scenario_config=ScenarioConfig(
        label="leaker",
        ScenarioCls=PassiveHijack,
        override_attacker_asns={CommonASNs.ATTACKER},
        override_legitimate_origin_asns={CommonASNs.LEGITIMATE_ORIGIN},
        # AS 1 and VICTIM use OnlyToCustomers
        override_base_settings={
            CommonASNs.ATTACKER: {Settings.LEAKER: True},
        },
    ),
    as_graph=ASGraph(graph_data),
    diagram_desc="Accidental route leak",
)



from .accidental_route_leak import AccidentalRouteLeak
from .first_asn_stripping_prefix_hijack import FirstASNStrippingPrefixHijack
from .forged_origin_prefix_hijack import ForgedOriginPrefixHijack
from .legitimate_prefix_only import LegitimatePrefixOnly
from .non_routed_prefix_hijack import NonRoutedPrefixHijack
from .non_routed_superprefix_hijack import NonRoutedSuperprefixHijack
from .non_routed_superprefix_prefix_hijack import NonRoutedSuperprefixPrefixHijack
from .passive_hijack import PassiveHijack
from .prefix_hijack import PrefixHijack
from .shortest_path_prefix_hijack import ShortestPathPrefixHijack
from .subprefix_hijack import SubprefixHijack
from .superprefix_prefix_hijack import SuperprefixPrefixHijack


__all__ = [
    "AccidentalRouteLeak",
    "FirstASNStrippingPrefixHijack",
    "ForgedOriginPrefixHijack",
    "LegitimatePrefixOnly",
    "NonRoutedPrefixHijack",
    "NonRoutedSuperprefixHijack",
    "NonRoutedSuperprefixPrefixHijack",
    "PassiveHijack",
    "PrefixHijack",
    "ShortestPathPrefixHijack",
    "SubprefixHijack",
    "SuperprefixPrefixHijack",
]



from .line_chart import LineChart
from .line_chart_factory import LineChartFactory
from .line import Line
from .line_properties_generator import LinePropertiesGenerator

__all__ = [
    "LineChart",
    "LineChartFactory",
    "Line",
    "LinePropertiesGenerator",
]



import dataclasses
from collections import UserDict
from typing import Any

from bgpsimulator.shared import Prefix, Relationships
from bgpsimulator.simulation_engine import Announcement as Ann


@dataclasses.dataclass(frozen=True, slots=True)
class AnnInfo:
    """Dataclass for storing a ribs in Ann info

    These announcements are unprocessed, so we store
    the unprocessed_ann and also the recv_relationship
    (since the recv_relationship on the announcement is
    from the last AS and has not yet been updated)
    """

    unprocessed_ann: "Ann"
    recv_relationship: "Relationships"

    def to_json(self) -> dict[str, Any]:
        """Returns a JSON representation of the AnnInfo"""

        return {
            "unprocessed_ann": self.unprocessed_ann.to_json(),
            "recv_relationship": self.recv_relationship,
        }

    @classmethod
    def from_json(cls, json: dict[str, Any]) -> "AnnInfo":
        """Returns an AnnInfo from a JSON representation"""

        return cls(
            unprocessed_ann=Ann.from_json(json["unprocessed_ann"]),
            recv_relationship=Relationships(json["recv_relationship"]),
        )


class AdjRIBsIn(UserDict[int, dict[Prefix, AnnInfo]]):
    """Incomming announcements for a BGP AS

    neighbor_asn: {prefix: (unprocessed_ann, relationship)}
    """

    def get_unprocessed_ann_recv_rel(
        self, neighbor_asn: int, prefix: Prefix
    ) -> AnnInfo | None:
        """Returns AnnInfo for a neighbor ASN and prefix"""

        return self.data.get(neighbor_asn, dict()).get(prefix)

    def add_unprocessed_ann(
        self,
        unprocessed_ann: "Ann",
        recv_relationship: "Relationships",
    ):
        """Adds an unprocessed ann to ribs in"""

        # Shorten the var name
        ann = unprocessed_ann
        if ann.as_path[0] not in self.data:
            self.data[ann.as_path[0]] = {
                ann.prefix: AnnInfo(
                    unprocessed_ann=unprocessed_ann, recv_relationship=recv_relationship
                )
            }
        else:
            self.data[ann.as_path[0]][ann.prefix] = AnnInfo(
                unprocessed_ann=unprocessed_ann, recv_relationship=recv_relationship
            )

    def get_ann_infos(self, prefix: Prefix) -> list[AnnInfo]:
        """Returns AnnInfos for a given prefix"""

        ann_infos = []
        for prefix_ann_info in self.data.values():
            ann_info = prefix_ann_info.get(prefix)
            if ann_info:
                ann_infos.append(ann_info)
        return ann_infos

    def remove_entry(self, neighbor_asn: int, prefix: Prefix):
        """Removes AnnInfo from RibsIn

        In real life, ASes ignore cases where withdrawals don't have a corresponding
        announcement, which is why we don't raise an error here
        """

        try:
            del self.data[neighbor_asn][prefix]
        except KeyError:
            pass

    def to_json(self) -> dict[int, dict[str, dict[str, Ann | Relationships]]]:
        """Returns a JSON representation of the AdjRIBsIn"""

        json_obj = {}
        for neighbor_asn, prefix_ann_info in self.data.items():
            json_obj[neighbor_asn] = {
                str(prefix): ann_info.to_json()
                for prefix, ann_info in prefix_ann_info.items()
            }
        return json_obj

    @classmethod
    def from_json(
        cls, json: dict[int, dict[str, dict[str, dict[str, Ann | Relationships]]]]
    ) -> "AdjRIBsIn":
        """Returns a AdjRIBsIn from a JSON representation"""

        adj_ribs_in = cls()
        for _neighbor_asn, prefix_ann_infos in json.items():
            for _prefix, ann_info_json in prefix_ann_infos.items():
                ann_info = AnnInfo.from_json(ann_info_json)
                adj_ribs_in.add_unprocessed_ann(
                    ann_info.unprocessed_ann, ann_info.recv_relationship
                )
        return adj_ribs_in



from typing import TYPE_CHECKING

from bgpsimulator.shared.enums import Relationships, Settings

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class ASPA:
    """A Policy that deploys ASPA and ASPA Records

    We experimented with adding a cache to the provider_check
    but this has a negligible impact on performance

    Removing the path reversals sped up performance by about 5%
    but made the code a lot less readable and deviated from the RFC,
    so we decided to forgo those as well

    This is now up to date with the V18 proposal
    """

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Returns False if ann from peer/customer when ASPA is set"""

        # Note: This first if check has to be removed if you want to implement
        # route server to RS-Client behaviour
        if not ASPA._next_hop_valid(policy, ann):
            return False
        # Most ASes recieve anns from providers (moved here for speed)
        elif from_rel == Relationships.PROVIDERS:
            return ASPA._downstream_check(policy, ann, from_rel)
        elif from_rel in (Relationships.CUSTOMERS, Relationships.PEERS):
            return ASPA._upstream_check(policy, ann, from_rel)
        else:
            raise NotImplementedError("Should never reach here")

    @staticmethod
    def _next_hop_valid(policy: "Policy", ann: "Ann") -> bool:
        """Ensures the next hop is the first ASN in the AS-Path

        Route servers are allowed to strip their own ASN (and in most cases are
        obligated to)
        """

        return ann.next_hop_asn == ann.as_path[0] or policy.as_.ixp

    @staticmethod
    def _upstream_check(
        policy: "Policy", ann: "Ann", from_rel: "Relationships"
    ) -> bool:
        """ASPA upstream check"""

        # Upstream check
        if len(ann.as_path) == 1:
            return True
        # For every adopting ASPA AS in the path,
        # The next ASN in the path must be in their providers list
        # Since this is checking from customers

        # 4. If max_up_ramp < N, the procedure halts with the outcome "Invalid".
        elif ASPA._get_max_up_ramp_length(policy, ann) < len(ann.as_path):
            return False

        # ASPA valid or unknown
        return True

    @staticmethod
    def _get_max_up_ramp_length(policy: "Policy", ann: "Ann") -> int:
        """See desc

        Determine the maximum up-ramp length as I, where I is the minimum
        index for which authorized(A(I), A(I+1)) returns "Not Provider+".  If
        there is no such I, the maximum up-ramp length is set equal to the
        AS_PATH length N.  This parameter is defined as max_up_ramp

        The up-ramp starts at AS(1) and each hop AS(i) to AS(i+1) represents
        Customer and Provider peering relationship. [i.e they reverse the path]
        """

        reversed_path = ann.as_path[::-1]

        for i in range(len(reversed_path) - 1):
            if not ASPA._provider_check(policy, reversed_path[i], reversed_path[i + 1]):
                return i + 1
        return len(ann.as_path)

    @staticmethod
    def _downstream_check(
        policy: "Policy", ann: "Ann", from_rel: "Relationships"
    ) -> bool:
        """ASPA downstream check"""

        # 4. If max_up_ramp + max_down_ramp < N,
        # the procedure halts with the outcome "Invalid".
        max_up_ramp = ASPA._get_max_up_ramp_length(policy, ann)
        max_down_ramp = ASPA._get_max_down_ramp_length(policy, ann)
        if max_up_ramp + max_down_ramp < len(ann.as_path):
            return False

        # ASPA Valid or Unknown (but not invalid)
        return True

    @staticmethod
    def _get_max_down_ramp_length(policy: "Policy", ann: "Ann") -> int:
        """See desc

        Similarly, the maximum down-ramp length can be determined as N - J +
        1 where J is the maximum index for which authorized(A(J), A(J-1))
        returns "Not Provider+".  If there is no such J, the maximum down-
        ramp length is set equal to the AS_PATH length N.  This parameter is
        defined as max_down_ramp.

        In the down-ramp, each pair AS(j) to
        AS(j-1) represents Customer and Provider peering relationship
        """

        reversed_path = ann.as_path[::-1]

        # We want the max J, so start at the end of the reversed Path
        # This is the most efficient way to traverse this
        for i in range(len(reversed_path) - 1, 0, -1):
            if not ASPA._provider_check(policy, reversed_path[i], reversed_path[i - 1]):
                # Must add one due to zero indexing in python, vs 1 indexing in RFC
                J = i + 1
                return len(reversed_path) - J + 1
        return len(ann.as_path)

    @staticmethod
    def _provider_check(policy: "Policy", asn1: int, asn2: int) -> bool:
        """Returns False if asn2 is not in asn1's provider_asns, AND asn1 adopts ASPA

        This also essentially can take the place of the "hop check" listed in
        ASPA RFC section 5 in ASPA v16
        or in ASPAv18 it takes the place of the provider auth func

        False indicates Not Provider+
        True indicates No Attestation or Provider+

        Updated so that if either AS doesn't exist, this function returns properly
        """

        cur_as_obj = policy.as_.as_graph.as_dict.get(asn1)
        if cur_as_obj and cur_as_obj.policy.settings[Settings.ASPA]:
            next_as_obj = policy.as_.as_graph.as_dict.get(asn2)
            next_asn = next_as_obj.asn if next_as_obj else next_as_obj
            if next_asn not in cur_as_obj.provider_asns:
                return False
        return True



from ipaddress import IPv6Network, ip_network

from .exceptions import ReservedPrefixError


class Prefix(IPv6Network):
    """Prefix class that is faster than ipaddress.ip_network

    All prefixes (IPv4 and IPv6) are internally stored as IPv6 addresses
    using IPv4-mapped IPv6 format (::ffff:a.b.c.d) for IPv4 addresses.
    """

    __slots__ = ("_hash",)

    def __init__(self, prefix: str, *args, **kwargs):
        """Create a fast Prefix from a string."""

        prefix = str(prefix)
        self._og_str_prefix = prefix
        temp_prefix = ip_network(prefix)
        if temp_prefix.is_reserved:
            raise ReservedPrefixError(
                f"Prefix {prefix} is reserved. Reserved prefixes can't be used "
                "since we map IPv4 prefixes to IPv6 prefixes."
            )
        if temp_prefix.version == 4:
            prefix = (
                f"::ffff:{temp_prefix.network_address}/{96 + temp_prefix.prefixlen}"
            )

        # Prefix is used as a key in dicts, so hash it in advance as it gets
        # called millions of times. We multiply the network address by 1000 to
        # avoid collisions with IPv6 prefix lengths (max 128)
        self._hash: int = hash(
            int(ip_network(prefix).network_address) * 1000 + temp_prefix.prefixlen
        )

        super().__init__(prefix, *args, **kwargs)

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other) -> bool:
        if isinstance(other, Prefix):
            return self._hash == other._hash
        else:
            return NotImplemented

    def __str__(self) -> str:
        return self._og_str_prefix

    def __repr__(self) -> str:
        return self._og_str_prefix



from typing import TYPE_CHECKING

from bgpsimulator.shared import Relationships, ROAValidity

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class ROV:
    """A Policy that deploys ROV"""

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Returns False if ann is ROV invalid"""
        (roa_validity, roa_routed) = policy.route_validator.get_roa_outcome(
            ann.prefix, ann.origin
        )
        # NOTE: Must work off of isinvalid, since Valid could be False but value could
        # be ROAValidity.UNKNOWN, which should not result in a reject.
        return not ROAValidity.is_invalid(roa_validity)



from .data_tracker import DataTracker, LineFilter
from .scenarios import (
    ScenarioConfig,
    Scenario,
    AccidentalRouteLeak,
    FirstASNStrippingPrefixHijack,
    ForgedOriginPrefixHijack,
    LegitimatePrefixOnly,
    NonRoutedPrefixHijack,
    NonRoutedSuperprefixHijack,
    NonRoutedSuperprefixPrefixHijack,
    PassiveHijack,
    PrefixHijack,
    ShortestPathPrefixHijack,
    SubprefixHijack,
    SuperprefixPrefixHijack,
)
from .simulation import Simulation
from .data_plane_packet_propagator import DataPlanePacketPropagator

__all__ = [
    "DataTracker",
    "LineFilter",
    "ScenarioConfig",
    "SubprefixHijack",
    "Scenario",
    "Simulation",
    "AccidentalRouteLeak",
    "FirstASNStrippingPrefixHijack",
    "ForgedOriginPrefixHijack",
    "LegitimatePrefixOnly",
    "NonRoutedPrefixHijack",
    "NonRoutedSuperprefixHijack",
    "NonRoutedSuperprefixPrefixHijack",
    "PassiveHijack",
    "PrefixHijack",
    "ShortestPathPrefixHijack",
    "SubprefixHijack",
    "SuperprefixPrefixHijack",
    "DataPlanePacketPropagator",
]



from typing import TYPE_CHECKING

from bgpsimulator.shared import Relationships

from .rov import ROV

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine.policy.policy import Policy
    from bgpsimulator.simulationengine import Announcement as Ann


class PeerROV:
    """A Policy that deploys Peer ROV (ROV only at peers)"""

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Returns False if ann is ROV invalid and is from a peer"""
        if from_rel == Relationships.PEERS:
            return ROV.valid_ann(policy, ann, from_rel)
        return True



from typing import TYPE_CHECKING

from bgpsimulator.shared import Relationships

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class PeerLockLite:
    """A Policy that deploys PeerLock Lite"""

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Returns False if ann is PeerLock Lite invalid"""

        as_dict = policy.as_.as_graph.as_dict
        if from_rel == Relationships.CUSTOMERS:
            # Tier-1 ASes have no providers, so if they are your customer,
            # there is a route leakage
            for asn in ann.as_path:
                if as_dict[asn].tier_1:
                    return False
        return True



from typing import TYPE_CHECKING

from bgpsimulator.shared import Relationships

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine.policy.policy import Policy
    from bgpsimulator.simulation_framework import Scenario


class AnnounceThenWithdraw:
    """In the second round of propagation, withdraw everything that was announced"""

    @staticmethod
    def pre_propagation_hook(
        policy: "Policy", propagation_round: int, scenario: "Scenario"
    ) -> None:
        """In the second round, withdraw everything you announced"""

        if propagation_round == 1:
            for ann in list(policy.local_rib.values()):
                # Withdraw everything that we announced
                if ann.recv_relationship == Relationships.ORIGIN:
                    withdraw_ann = policy.local_rib.pop(ann.prefix).copy(withdraw=True)
                    policy.withdraw_ann_from_neighbors(withdraw_ann)



from functools import lru_cache

from bgpsimulator.shared import Prefix, ROARouted, ROAValidity

from .roa import ROA
from .roas_node import ROASNode


class RouteValidator:
    """Trie of ROAs"""

    def __init__(self) -> None:
        self.root: ROASNode = ROASNode()
        self.get_roa_outcome.cache_clear()
        self._roas: list[ROA] = list()

    @property
    def roas(self) -> tuple[ROA, ...]:
        """Returns all ROAs in the trie"""
        return tuple(self._roas)

    def clear(self) -> None:
        """Clears the trie"""
        self.__init__()  # type: ignore

    def add_roa(self, roa: ROA) -> None:
        """Inserts a prefix into the trie"""
        prefix = roa.prefix
        bits = self._get_binary_str_from_prefix(prefix)
        node = self.root
        for bit in bits[: prefix.prefixlen]:
            if bool(int(bit)):
                if node.right is None:
                    node.right = ROASNode()
                node = node.right
            else:
                if node.left is None:
                    node.left = ROASNode()
                node = node.left
        node.add_data(prefix, roa)
        self._roas.append(roa)

    def __contains__(self, prefix: Prefix) -> bool:
        """Checks if a prefix is contained within the Trie"""
        return bool(self.get_most_specific_trie_supernet(prefix))

    def get_most_specific_trie_supernet(self, prefix: Prefix) -> ROASNode | None:
        """Returns the most specific trie subnet"""
        bits = self._get_binary_str_from_prefix(prefix)
        node = self.root
        most_specific_node = None
        for bit in bits[: prefix.prefixlen]:
            next_node = node.right if bool(int(bit)) else node.left
            if next_node is None:
                return most_specific_node
            elif next_node.prefix is not None:
                most_specific_node = next_node
            node = next_node
        return most_specific_node

    def _get_binary_str_from_prefix(self, prefix: Prefix) -> str:
        """Returns a binary string from a prefix"""

        binary_str = ""
        for _byte in prefix.network_address.packed:
            binary_str += str(bin(_byte))[2:].zfill(8)
        return binary_str

    @lru_cache(maxsize=10_000)  # noqa: B019
    def get_roa_outcome(
        self, prefix: Prefix, origin: int
    ) -> tuple[ROAValidity, ROARouted]:
        """Gets the validity and roa routed vs non rotued of a prefix-origin pair

        This can get fairly complicated, since there can be multiple ROAs
        for the same announcement, and each ROA can have a different validity.
        Essentially, we need to calculate the best validity for a given announcement
        and then that's the validity that should be used. I.e. the "most valid" roa
        is the one that should be used
        """

        relevant_roas = self.get_relevant_roas(prefix)

        if relevant_roas:
            # Return the best ROAOutcome
            rv = sorted(
                [x.get_outcome(prefix, origin) for x in relevant_roas],
                key=lambda x: x[0],
            )[0]
            return rv
        else:
            return ROAValidity.UNKNOWN, ROARouted.UNKNOWN

    def get_relevant_roas(self, prefix: Prefix) -> set[ROA]:
        """Returns all relevant ROAs for a given prefix

        This is a bit non-intuitive, because you'd think that all ROAs relevant
        for a given prefix would be __at the prefix__. But this isn't the case.
        They could be higher up the trie, and simply have a __max length__ that
        is >= your prefix length. So you need to collect all ROAs all the way down,
        and for each ROA, determine if it's relevant, and then return them
        """

        bits = self._get_binary_str_from_prefix(prefix)
        node = self.root
        roas: list[ROA] = list()
        for bit in bits[: prefix.prefixlen]:
            next_node = node.right if bool(int(bit)) else node.left
            if next_node is None:
                return set(roas)
            elif next_node.prefix is not None:
                for roa in next_node.roas:
                    if roa.covers_prefix(prefix):
                        roas.append(roa)
            node = next_node
        return set(roas)



import warnings
from collections import deque

from bgpsimulator.route_validator import ROA
from bgpsimulator.shared import ASNGroups, IPAddr, bgpsimulator_logger
from bgpsimulator.shared.enums import (
    CommonPrefixes,
    Relationships,
    Timestamps,
)
from bgpsimulator.simulation_engine import Announcement as Ann
from bgpsimulator.simulation_engine import SimulationEngine
from bgpsimulator.simulation_framework.scenarios.scenario import Scenario


class AccidentalRouteLeak(Scenario):
    """Attacker leaks anns received from peers/providers"""

    min_propagation_rounds: int = 2

    def __init__(self, *args, **kwargs):
        self._attackers_customer_cone_asns: set[int] = set()
        super().__init__(*args, **kwargs)
        self._validate_attacker_asn_group()

    def _validate_attacker_asn_group(self):
        """Validates that the attacker is in an ASN group that can leak"""

        if (
            self.scenario_config.attacker_asn_group in self.warning_asn_groups
            and not self.scenario_config.override_attacker_asns
        ):
            msg = (
                "You used the ASNGroup of "
                f"{self.scenario_config.attacker_asn_group} "
                f"for your scenario {self.__class__.__name__}, "
                f"but {self.__class__.__name__} can't leak from stubs. "
                "To suppress this warning, override warning_as_groups. "
                "To change the ASNGroup to something other than stubs, you can "
                " set attacker_asn_group=ASNGroups.MULTIHOMED.value, "
                " in the scenario config after importing like "
                "from bgpsimulator.shared import ASNGroups"
            )
            warnings.warn(msg, RuntimeWarning, stacklevel=2)

    @property
    def warning_asn_groups(self) -> frozenset[str]:
        """Returns a frozenset of ASNGroups that should raise a warning"""

        return frozenset(
            [
                ASNGroups.STUBS_OR_MH.value,
                ASNGroups.STUBS.value,
                ASNGroups.ALL_WOUT_IXPS.value,
            ]
        )

    def post_propagation_hook(
        self,
        engine: "SimulationEngine",
        percent_ases_randomly_adopting: float,
        trial: int,
        propagation_round: int,
    ) -> None:
        """Causes an accidental route leak

        Changes the valid prefix to be received from a customer
        so that in the second propagation round, the AS will export to all
        relationships

        NOTE: the old way of doing this was to simply alter the attackers
        local RIB and then propagate again. However - this has some drawbacks
        Then the attacker must deploy BGPFull (that uses withdrawals) and
        the entire graph has to propagate again. BGPFull (and subclasses
        of it) are MUCH slower than BGP due to all the extra
        computations for withdrawals, AdjRIBsIn, AdjRIBsOut, etc. Additionally,
        propagating a second round after the ASGraph is __already__ full
        is wayyy more expensive than propagating when the AS graph is empty.

        Instead, we now get the announcement that the attacker needs to leak
        after the first round of propagating the valid prefix.
        Then we clear the graph, seed those announcements, and propagate again
        This way, we avoid needing BGPFull (since the graph has been cleared,
        there is no need for withdrawals), and we avoid propagating a second
        time after the graph is alrady full.

        Since this simulator treats each propagation round as if it all happens
        at once, this is possible.

        Additionally, you could also do the optimization in the first propagation
        round to only propagate from ASes that can reach the attacker. But we'll
        forgo this for now for simplicity.
        """

        if propagation_round == 0:
            seed_asn_ann_dict: dict[int, list[Ann]] = self.seed_asn_ann_dict.copy()
            for attacker_asn in self.attacker_asns:
                if not engine.as_graph.as_dict[attacker_asn].policy.local_rib:
                    bgpsimulator_logger.warning(
                        "Attacker did not recieve announcement, can't leak."
                    )
                for _prefix, ann in engine.as_graph.as_dict[
                    attacker_asn
                ].policy.local_rib.items():
                    seed_asn_ann_dict.setdefault(attacker_asn, []).append(
                        ann.copy(
                            recv_relationship=Relationships.ORIGIN,
                            timestamp=Timestamps.ATTACKER.value,
                        )
                    )
            self.seed_asn_ann_dict = seed_asn_ann_dict
            self.setup_engine(engine)
        elif propagation_round > 1:
            raise NotImplementedError

    def _get_seed_asn_ann_dict(self, engine: SimulationEngine) -> dict[int, list[Ann]]:
        anns = dict()
        for legitimate_origin_asn in self.legitimate_origin_asns:
            anns[legitimate_origin_asn] = [
                Ann(
                    prefix=CommonPrefixes.PREFIX.value,
                    as_path=(legitimate_origin_asn,),
                    next_hop_asn=legitimate_origin_asn,
                    recv_relationship=Relationships.ORIGIN,
                    timestamp=Timestamps.LEGITIMATE_ORIGIN,
                )
            ]
        return anns

    def _get_roas(
        self,
        *,
        seed_asn_ann_dict: dict[int, list[Ann]],
        engine: SimulationEngine,
    ) -> list[ROA]:
        """Returns a list of ROAs"""

        return [
            ROA(CommonPrefixes.PREFIX.value, x) for x in self.legitimate_origin_asns
        ]

    def _get_dest_ip_addr(self) -> IPAddr:
        """Returns the destination IP address for the scenario"""

        return IPAddr("1.2.3.4")

    @property
    def untracked_asns(self) -> set[int]:
        """Returns ASNs that shouldn't be tracked by the data tracker

        By default just the default adopters and non adopters
        however for the route leak, we don't want to track the customers of the
        leaker, since you can not "leak" to your own customers
        """

        return super().untracked_asns | self._attackers_customer_cone_asns

    def _get_attacker_asns(
        self,
        override_attacker_asns: set[int] | None,
        prev_attacker_asns: set[int] | None,
        engine: SimulationEngine,
    ) -> set[int]:
        """Gets attacker ASNs, overriding the valid prefix which has no attackers

        There is a very rare case where the attacker can not perform the route leak
        due to a disconnection, which happens around .1% in the CAIDA topology.
        In theory - you could just look at the provider cone of the victim,
        and then the peers of that provider cone (and of the victim itself), and
        then the customer cones of all of those ASes to get the list of possible
        valid attackers. However, we consider the attacker being unable to attack
        in extremely rare cases a valid result, and thus do not change the random
        selection. Doing so would also be a lot slower for a very extreme edge case
        """

        attacker_asns = super()._get_attacker_asns(
            override_attacker_asns, prev_attacker_asns, engine
        )
        # Add customer cones so that we can avoid them when tracking data
        for attacker_asn in attacker_asns:
            self._attackers_customer_cone_asns.update(
                self._get_customer_cone_asns(attacker_asn, engine)
            )
        return attacker_asns

    def _get_customer_cone_asns(self, asn: int, engine: SimulationEngine) -> set[int]:
        """Returns the customer cone of an AS"""

        customer_cone_asns: set[int] = set()
        fifo_queue: deque[int] = deque([asn])
        while fifo_queue:
            asn = fifo_queue.popleft()
            customer_cone_asns.update(engine.as_graph.as_dict[asn].customer_asns)
            fifo_queue.extend(engine.as_graph.as_dict[asn].customer_asns)
        return customer_cone_asns

    def _get_possible_legitimate_origin_asns(
        self,
        engine: SimulationEngine,
        percent_ases_randomly_adopting: float,
    ) -> set[int]:
        """Returns possible legitimate_origin ASNs, defaulted from config

        Removes attacker's customer cones from possible victims (or else
        there would be no leakage)
        """

        possible_asns = super()._get_possible_legitimate_origin_asns(
            engine, percent_ases_randomly_adopting
        )
        # Remove attacker's customer conesfrom possible victims
        possible_asns = possible_asns.difference(self._attackers_customer_cone_asns)
        return possible_asns



from bgpsimulator.as_graphs.as_graph import ASGraph
from bgpsimulator.engine_runner import EngineRunConfig
from bgpsimulator.shared import CommonASNs, Settings
from bgpsimulator.simulation_framework import PrefixHijack, ScenarioConfig
from bgpsimulator.shared.enums import CommonPrefixes, Relationships, Timestamps
from bgpsimulator.simulation_engine import Announcement as Ann


# Graph to test OTC from a peer
# CommonASNs.LEGITIMATE_ORIGIN - CommonASNs.ATTACKER - 1
graph_data = {
    "ases": {
        "1": {"asn": 1, "customer_asns": [CommonASNs.ATTACKER]},
        CommonASNs.ATTACKER: {
            "asn": CommonASNs.ATTACKER,
            "customer_asns": [CommonASNs.LEGITIMATE_ORIGIN],
            "provider_asns": [1],
        },
        CommonASNs.LEGITIMATE_ORIGIN: {
            "asn": CommonASNs.LEGITIMATE_ORIGIN,
            "provider_asns": [CommonASNs.ATTACKER],
        },
    }
}

# Create the engine run config
ex_config_034 = EngineRunConfig(
    name="ex_034_withdrawal_supppresion_loop",
    scenario_config=ScenarioConfig(
        label="withdrawal_suppression_loop",
        ScenarioCls=PrefixHijack,
        override_seed_asn_ann_dict={
            CommonASNs.LEGITIMATE_ORIGIN: [
                Ann(
                    prefix=CommonPrefixes.PREFIX.value,
                    as_path=(CommonASNs.LEGITIMATE_ORIGIN,),
                    next_hop_asn=CommonASNs.LEGITIMATE_ORIGIN,
                    recv_relationship=Relationships.ORIGIN,
                    timestamp=Timestamps.LEGITIMATE_ORIGIN,
                ),
            ],
            1: [
                Ann(
                    prefix=CommonPrefixes.SUBPREFIX.value,
                    as_path=(1,),
                    next_hop_asn=1,
                    recv_relationship=Relationships.ORIGIN,
                    timestamp=Timestamps.LEGITIMATE_ORIGIN,
                ),
            ],
        },
        num_attackers=0,
        override_attacker_asns=set(),  # CommonASNs.ATTACKER},
        override_legitimate_origin_asns={CommonASNs.LEGITIMATE_ORIGIN},
        # AS 1 and VICTIM use OnlyToCustomers
        override_base_settings={
            1: {Settings.ANNOUNCE_THEN_WITHDRAW: True},
            CommonASNs.ATTACKER: {
                Settings.NEVER_PROPAGATE_WITHDRAWALS: True,
                Settings.NEVER_WITHDRAW: True,
            },
        },
    ),
    as_graph=ASGraph(graph_data),
    diagram_desc="Withdrawal suppression loop",
)



from typing import TYPE_CHECKING

from bgpsimulator.shared.enums import Relationships, Settings

from .aspa import ASPA

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class ASRA:
    """Algo-B using ASRA3 records"""

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """
        1) Perform standard ASPA route-leak detection via ASPA.valid_ann()
           If ASPA.valid_ann() returns False => 'Invalid'. We bail out.
        2) If ASPA.valid_ann() returns True => route is either 'Valid' or 'Unknown'.
           If from_rel == PROVIDER, we run the extra ASRA-B check
             - We figure out min_up_ramp
             - For each hop from min_up_ramp to the end of the path, check
               if it's a fake link.
           Otherwise, we do nothing more.
        """

        # 1) Run ASPA's logic
        aspa_result = ASPA.valid_ann(policy, ann, from_rel)

        # If ASPA deems it invalid, bail
        if not aspa_result:
            return False

        # 2) If from_rel is not PROVIDER, do nothing more
        if from_rel != Relationships.PROVIDERS:
            return True

        # If from_rel == PROVIDER, do ASRA-B "fake link" checks
        # i.e., check from min_up_ramp up to the end of the path

        path = ann.as_path[::-1]
        n = len(path)

        # Compute min_up_ramp in a simple way:
        # We start from the origin side (index 0) going forward until
        # we find the first place the ASPA "provider check" fails.
        min_up_ramp = ASRA._get_min_up_ramp_length(policy, ann)

        # If min_up_ramp == n, that means every hop was "provider+",
        # so there's no leftover "down-ramp" region to apply ASRA to,
        # and the route remains valid or unknown => accept it
        if min_up_ramp == n:
            return True

        # Else, check each hop from i=min_up_ramp..(n-2),
        # i.e., the region from the top of the up-ramp to the second-last
        # AS in the path.  (We examine i -> i+1.)
        for i in range(min_up_ramp, n - 1):
            if ASRA._is_fake_link(policy, path[i], path[i + 1]):
                # RFC: "If Fake-Link(...) = Detected, then outcome => invalid"
                return False

        # If no fake link discovered, keep it
        return True

    @staticmethod
    def _get_min_up_ramp_length(policy: "Policy", ann: "Ann") -> int:
        """
        We define min_up_ramp ~ the first 'i' from the origin side
        where the ASPA check fails or the AS does not adopt ASPA.

        If we never fail, we return len(path).
        """
        path = ann.as_path[::-1]
        for i in range(len(path) - 1):
            asn1 = path[i]
            asn2 = path[i + 1]
            asn1_obj = policy.as_.as_graph.as_dict.get(asn1)

            # 1/5/2024 JF: Added check for if as1_obj doesn't exist
            # If asn1 does not adopt ASPA, we treat that
            # as 'No Attestation', so min_up_ramp ends here.
            if not asn1_obj or not asn1_obj.policy.settings[Settings.ASPA]:
                return i

            # If asn2 is not in asn1's provider list => 'Not Provider+',
            # so min_up_ramp ends here
            if asn2 not in asn1_obj.provider_asns:
                return i

        return len(path)

    @staticmethod
    def _is_fake_link(policy: "Policy", asn1: int, asn2: int) -> bool:
        """
        ASRA-B fake link check (Algorithm B).
        "If AS(i) has valid ASPA(s) and does not list AS(i+1) as a provider,
         AND AS(i) has valid ASRA(s) and does not list AS(i+1) as neighbor,
         => FAKE LINK."

        We'll interpret "valid ASPA(s)" as "this AS is adopting ASPA
        (i.e. policy is ASPA or child class), and asn2 is not in
        asn1.provider_asns."

        We'll interpret "valid ASRA(s)" as "this AS is adopting ASRA
        (i.e. policy is ASRA or child class), and asn2 is not in
        asn1.neighbor_asns."
        """
        asn1_obj = policy.as_.as_graph.as_dict.get(asn1)

        # Must meet BOTH conditions to declare fake link:
        # 1) asn1 adopts ASPA and does NOT list asn2 as a provider
        has_aspa_but_not_provider = bool(
            asn1_obj
            and asn1_obj.policy.settings[Settings.ASPA]
            and asn2 not in asn1_obj.provider_asns
        )

        # 2) asn1 also adopts ASRA and does NOT list asn2 as neighbor
        has_asra_but_not_neighbor = bool(
            asn1_obj
            and asn1_obj.policy.settings[Settings.ASRA]
            and asn2 not in asn1_obj.neighbor_asns
        )

        # If both are True => fake link
        return has_aspa_but_not_provider and has_asra_but_not_neighbor



from typing import TYPE_CHECKING

from bgpsimulator.shared import Relationships, Settings

from .aspa import ASPA

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class ASPAwN:
    """ASRA: Esentially ASPA and checking neighbors at every AS together

    NOTE: Updated 12/26/2024. This is the same as ASRA algo B, confirmed

    Originally we had this policy coded for ASPAwN, an extension of ASPA that
    we had developed independently of ASRA in parallel. Upon publishing an ASPA
    evaluation paper, we received feedback from an ASPA RFC author and converted
    ASPAwN into ASRA

    That being said, ASRA is not yet finalized, and at the time of this writing the
    draft wasn't even published. This policy assumes ASRA3 records (not specifying
    the neighbor type) and also assumes the strict algorithm.

    As ASPAwN is equivalent to ASRA, this is merely vestigial and I would recommend
    using the ASRA policy. However - ASRA is the same as ASPAwN and imo this is a
    vastly more simple algorithm.
    """

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Checks neighbors at every AS in the AS-Path"""

        as_path = ann.as_path
        as_dict = policy.as_.as_graph.as_dict
        for i, asn in enumerate(as_path):
            # Get the AS object for the current AS in the AS Path
            asra_as_obj = as_dict.get(asn)
            # If the AS is an ASRA AS
            if asra_as_obj and asra_as_obj.policy.settings[Settings.ASPA_W_N]:
                # Check that both of it's neighbors are in the valid next hops
                for neighbor_index in (i - 1, i + 1):
                    # Can't use try except IndexError here, since -1 is a valid index
                    if 0 <= neighbor_index < len(as_path):
                        neighbor_asn = as_path[neighbor_index]
                        if neighbor_asn not in asra_as_obj.neighbor_asns:
                            return False
        # This is a superset of ASPA
        return ASPA.valid_ann(policy, ann, from_rel)



from typing import Any

from .as_graph_utils import ASGraphUtils
from .base_as import AS


class ASGraph:
    """BGP Topology"""

    def __eq__(self, other) -> bool:
        if isinstance(other, ASGraph):
            return self.as_dict == other.as_dict
        else:
            return NotImplemented

    ##############
    # Init Funcs #
    ##############

    def __init__(
        self,
        graph_data: dict[str, Any],
    ) -> None:
        """Reads in relationship data from a JSON and generate graph"""

        # Always add cycles, provider cones, and propagation ranks
        # if it hasn't been done already
        ASGraphUtils.add_extra_setup(graph_data)
        # populate basic info
        self.as_dict = {
            int(asn): AS.from_json(info, as_graph=self)
            for asn, info in graph_data["ases"].items()
        }
        # Populate ASN groups
        self.asn_groups = {
            asn_group_key: {int(x) for x in asn_group}
            for asn_group_key, asn_group in graph_data["asn_groups"].items()
        }
        # populate objects
        self._populate_objects()
        # Add propagation ranks
        self.propagation_ranks: list[list[AS]] = [
            [self.as_dict[int(asn)] for asn in rank]
            for rank in graph_data["propagation_rank_asns"]
        ]

    def _populate_objects(self) -> None:
        """Populates the AS objects with the relationships"""
        for _asn, as_obj in self.as_dict.items():
            as_obj.set_relations()

    ##################
    # Iterator funcs #
    ##################

    # https://stackoverflow.com/a/7542261/8903959
    def __getitem__(self, asn: int) -> AS:
        return self.as_dict[asn]

    def __iter__(self):
        return iter(self.as_dict.values())

    def __len__(self) -> int:
        return len(self.as_dict)

    ##############
    # JSON funcs #
    ##############

    def to_json(self) -> dict[str, Any]:
        """Converts the ASGraph to a JSON object"""

        return {
            "ases": {asn: as_obj.to_json() for asn, as_obj in self.as_dict.items()},
            "asn_groups": {
                asn_group_key: sorted(asn_group)
                for asn_group_key, asn_group in self.asn_groups.items()
            },
            "extra_setup_complete": True,
            "cycles_detected": False,
            "propagation_rank_asns": [
                [x.asn for x in rank] for rank in self.propagation_ranks
            ],
        }

    @classmethod
    def from_json(cls, json_obj: dict[str, Any]) -> "ASGraph":
        """Converts the ASGraph to a JSON object"""

        # Convert back to sets
        json_obj["asn_groups"] = {
            asn_group_key: {int(x) for x in asn_group}
            for asn_group_key, asn_group in json_obj.get("asn_groups", dict()).items()
        }
        json_obj["ases"] = {
            int(asn): as_json for asn, as_json in json_obj.get("ases", dict()).items()
        }
        json_obj["propagation_rank_asns"] = [
            [int(x) for x in rank] for rank in json_obj.get("propagation_rank_asns", [])
        ]
        return cls(json_obj)



from .policy import Policy
from .policy_extensions import (
    ASPathEdgeFilter,
    ASPA,
    ASPAwN,
    ASRA,
    BGP,
    OnlyToCustomers,
    EnforceFirstAS,
    ROV,
    PathEnd,
    PeerLockLite,
    ROVPPV1Lite,
    ROVPPV2Lite,
    ROVPPV2iLite,
    BGPSec,
    BGPiSecTransitive,
    ProviderConeID,
    OriginPrefixHijackCustomers,
    FirstASNStrippingPrefixHijackCustomers,
    PeerROV,
    ASPAPP,
    ROST,
    RoSTTrustedRepository,
    NeverPropagateWithdrawals,
    Leaker,
    AnnounceThenWithdraw,
)

__all__ = [
    "Policy",
    "ASPathEdgeFilter",
    "ASPA",
    "ASPAwN",
    "ASRA",
    "BGP",
    "OnlyToCustomers",
    "EnforceFirstAS",
    "ROV",
    "PathEnd",
    "PeerLockLite",
    "ROVPPV1Lite",
    "ROVPPV2Lite",
    "ROVPPV2iLite",
    "BGPSec",
    "BGPiSecTransitive",
    "ProviderConeID",
    "OriginPrefixHijackCustomers",
    "FirstASNStrippingPrefixHijackCustomers",
    "PeerROV",
    "ASPAPP",
    "ROST",
    "RoSTTrustedRepository",
    "NeverPropagateWithdrawals",
    "Leaker",
    "AnnounceThenWithdraw",
]



from math import sqrt
from statistics import stdev
from typing import Any

from bgpsimulator.shared import Outcomes
from bgpsimulator.simulation_engine import SimulationEngine
from bgpsimulator.simulation_framework.scenarios.scenario import Scenario

from .line_filter import LineFilter


class DataTracker:
    """Tracks data for later use in creating line charts"""

    __slots__ = (
        "line_filters",
        "scenario_labels",
        "percent_ases_randomly_adopting",
        "unaggregated_data",
        "aggregated_data",
    )

    def __init__(
        self,
        line_filters,
        scenario_labels,
        percent_ases_randomly_adopting,
        unaggregated_data=None,
        aggregated_data=None,
    ) -> None:
        """Format of the unaggregated_data below


        You have the scenario label as the outer key, representing the
        scenario config (who is attacking, defending, adopting, what routing
        policy settings they are using).
        Then for each scenario config/label, you have filters for a line
        (as the next key), using a subset of the data from the
        scenario_config.
        Within that there are the data points for each percent of ASes
        adopting, and their values.
        {
            # Name of the line from the scenario config
            scenario_label: {
                # The subset of data used from the scenario_config to create this line
                line_filter (as_group, in_adopting_asns, prop_round, outcome): {
                    # Data point in the line
                    percent_ases_randomly_adopting: [
                        {
                            # Numerator = achieved outcome of graph filter
                            # and within the filter
                            numerator: 0,
                            # Denominator = within the filter, disregarding outcome
                            denominator: 0
                        }
                    ]
                }
            }
        }

        In aggregated_data (which is the final output), a list of
        numerators and denominators -> averagevalue, yerr (90% confidence
        interval)

        This data structure has gone through many iterations.
        At first, it was a gigantic nested dictionary.
        Then it become a multi nested dictionary using Python classes,
        which were hard to convert to JSON.
        I think this will be the most usable, as it will be pure JSON, not
        a bunch of dataclasses
        and from a high level it makes sense.
        """
        self.line_filters = line_filters
        self.scenario_labels = scenario_labels
        self.percent_ases_randomly_adopting = percent_ases_randomly_adopting
        self.unaggregated_data = (
            unaggregated_data or self._create_new_unaggregated_data()
        )
        self.aggregated_data = (
            aggregated_data
            or {
                label: {line_filter: {} for line_filter in line_filters}
                for label in self.scenario_labels
            }
            or {}
        )

    def _create_new_unaggregated_data(
        self,
    ) -> dict[str, dict[LineFilter, dict[float, list[dict[str, int]]]]]:
        """Creates a new unaggregated data dictionary"""

        data: dict[str, dict[LineFilter, dict[float, list[dict[str, int]]]]] = dict()
        # This is the label of the scenario_config
        for label in self.scenario_labels:
            data[label] = dict()
            # These are the filters for the scenario_config
            # to create the line seen in the graph
            for line_filter in self.line_filters:
                data[label][line_filter] = dict()
                for percent_adopt in self.percent_ases_randomly_adopting:
                    data[label][line_filter][percent_adopt] = list()
        return data

    def __add__(self, other) -> "DataTracker":
        if isinstance(other, DataTracker):
            # Timing trials show this is the fastest way
            new_data = self._create_new_unaggregated_data()
            for label, inner_dict in self.unaggregated_data.items():
                for line_filter, trial_data in inner_dict.items():
                    for (
                        percent_ases_randomly_adopting,
                        _data_points,
                    ) in trial_data.items():
                        new_data[label][line_filter][percent_ases_randomly_adopting] = (
                            self.unaggregated_data[label][line_filter][
                                percent_ases_randomly_adopting
                            ]
                            + other.unaggregated_data[label][line_filter][
                                percent_ases_randomly_adopting
                            ]
                        )
            return DataTracker(
                self.line_filters,
                self.scenario_labels,
                self.percent_ases_randomly_adopting,
                new_data,
            )
        else:
            return NotImplemented

    def store_trial_data(
        self,
        *,
        engine: SimulationEngine,
        scenario: Scenario,
        asn_to_packet_outcome_dict: dict[int, Outcomes],
        propagation_round: int,
    ) -> None:
        """Stores the data for a trial"""

        # Initialize the trial data
        for _label, inner_dict in self.unaggregated_data.items():
            for _line_filter, trial_data in inner_dict.items():
                trial_data.setdefault(
                    scenario.percent_ases_randomly_adopting, []
                ).append({"numerator": 0, "denominator": 0})

        unaggregated_scenario_data = self.unaggregated_data[
            scenario.scenario_config.label
        ]

        for line_filter in unaggregated_scenario_data:
            # We don't need this (since this is also checked in
            # LineFilter), but it saves some time
            if line_filter.prop_round != propagation_round:
                continue
            for as_obj in engine.as_graph:
                # Don't count ASes that are preset, such as attackers, victims, etc.
                if as_obj.asn in scenario.untracked_asns:
                    continue
                outcome = asn_to_packet_outcome_dict[as_obj.asn]
                if line_filter.as_in_denominator(
                    as_obj, engine.as_graph, scenario, propagation_round, outcome
                ):
                    # The default for the numerator simply checks that the
                    # outcome is the same as the line filter
                    if line_filter.as_in_numerator(
                        as_obj, engine.as_graph, scenario, propagation_round, outcome
                    ):
                        unaggregated_scenario_data[line_filter][
                            scenario.percent_ases_randomly_adopting
                        ][-1]["numerator"] += 1
                    unaggregated_scenario_data[line_filter][
                        scenario.percent_ases_randomly_adopting
                    ][-1]["denominator"] += 1

    def aggregate_data(self) -> None:
        """Aggregates the data"""

        for label, inner_dict in self.unaggregated_data.items():
            for line_filter, trial_data in inner_dict.items():
                for percent_ases_randomly_adopting, data_points in trial_data.items():
                    decimal_vals = [
                        x["numerator"] * 100 / x["denominator"]
                        if x["denominator"] != 0
                        else 0
                        for x in data_points
                    ]
                    self.aggregated_data[label][line_filter][
                        percent_ases_randomly_adopting
                    ] = {
                        "value": sum(decimal_vals) / len(decimal_vals),  # average
                        "yerr": self._get_yerr(decimal_vals),
                    }
        # Clear the unaggregated data to save memory
        self.unaggregated_data = {}

    def _get_yerr(self, percent_list: list[float]) -> float:
        """Returns 90% confidence interval for graphing"""

        if len(percent_list) > 1:
            return 1.645 * 2 * stdev(percent_list) / sqrt(len(percent_list))
        else:
            return 0

    def to_json(self) -> dict[str, Any]:
        """Converts the data to a JSON-friendly format"""
        json_data: dict[str, dict[str, dict[str, int]]] = {}
        for label, inner_dict in self.aggregated_data.items():
            json_data[label] = {}
            for line_filter, trial_data in inner_dict.items():
                json_data[label][line_filter.to_json()] = trial_data
        return {
            "line_filters": [x.to_json() for x in self.line_filters],
            "scenario_labels": self.scenario_labels,
            "percent_ases_randomly_adopting": self.percent_ases_randomly_adopting,
            "aggregated_data": json_data,
            "schema_description": DataTracker.__init__.__doc__,
        }

    @classmethod
    def from_json(cls, json_data: dict[str, Any]) -> "DataTracker":
        """Converts the data from a JSON-friendly format"""
        aggregated_data: dict[str, dict[LineFilter, dict[str, int]]] = {}
        for label, inner_dict in json_data["aggregated_data"].items():
            aggregated_data[label] = {}
            for line_filter, trial_data in inner_dict.items():
                aggregated_data[label][LineFilter.from_json(line_filter)] = trial_data
        return cls(
            line_filters=[LineFilter.from_json(x) for x in json_data["line_filters"]],
            scenario_labels=json_data["scenario_labels"],
            percent_ases_randomly_adopting=json_data["percent_ases_randomly_adopting"],
            aggregated_data=aggregated_data,
        )

    def to_csv(self) -> str:
        """Converts the data to a CSV-friendly format"""

        csv_data = (
            "scenario_label,as_group,in_adopting_asns,prop_round,outcome,"
            "percent_ases_randomly_adopting,value,yerr,line_filter_json\n"
        )
        for label, inner_dict in self.aggregated_data.items():
            for line_filter, trial_data in inner_dict.items():
                for percent_ases_randomly_adopting, data_point in trial_data.items():
                    csv_data += (
                        f"{label},{line_filter.to_csv()},"
                        f"{percent_ases_randomly_adopting},"
                        f"{data_point['value']},{data_point['yerr']},"
                        f"{line_filter.to_json()}\n"
                    )
        return csv_data



from . import (  # isort: skip
    shared,
    simulation_engine,
    as_graphs,
    simulation_framework,
    route_validator,
    engine_runner,
)


__all__ = [
    "shared",
    "as_graphs",
    "simulation_engine",
    "simulation_framework",
    "route_validator",
    "engine_runner",
]



from typing import Any

from .line_properties_generator import LinePropertiesGenerator

GENERATOR = LinePropertiesGenerator()


class Line:
    """A line in a line chart"""

    def __init__(
        self,
        label: str,
        xs: list[float],
        ys: list[float],
        yerrs: list[float],
        marker: str = "",
        ls: str = "",
        color: str = "",
    ) -> None:
        self.label: str = label
        self.xs: list[float] = xs
        self.ys: list[float] = ys
        self.yerrs: list[float] = yerrs
        self.marker: str = marker
        self.ls: str = ls
        self.color: str = color

        if not self.marker:
            self.marker = GENERATOR.get_marker()
        if not self.ls:
            self.ls = GENERATOR.get_line_style()
        if not self.color:
            self.color = GENERATOR.get_color()

    def to_json(self) -> dict[str, Any]:
        return {
            "label": self.label,
            "xs": self.xs,
            "ys": self.ys,
            "yerrs": self.yerrs,
            "marker": self.marker,
            "ls": self.ls,
            "color": self.color,
        }

    @classmethod
    def from_json(cls, json_data: dict[str, Any]) -> "Line":
        return cls(
            label=json_data["label"],
            xs=[float(x) for x in json_data["xs"]],
            ys=[float(y) for y in json_data["ys"]],
            yerrs=[float(yerr) for yerr in json_data["yerrs"]],
            marker=json_data["marker"],
            ls=json_data["ls"],
            color=json_data["color"],
        )



from bgpsimulator.route_validator import ROA
from bgpsimulator.shared import IPAddr
from bgpsimulator.shared.enums import CommonPrefixes, Relationships, Timestamps
from bgpsimulator.simulation_engine import Announcement as Ann
from bgpsimulator.simulation_engine import SimulationEngine

from .shortest_path_prefix_hijack import (
    ShortestPathPrefixHijack,
)


class FirstASNStrippingPrefixHijack(ShortestPathPrefixHijack):
    """Attacker announces a prefix with the first ASN stripped"""

    def _get_seed_asn_ann_dict(self, engine: SimulationEngine) -> dict[int, list[Ann]]:
        seed_asn_ann_dict = dict()
        for legitimate_origin_asn in self.legitimate_origin_asns:
            seed_asn_ann_dict[legitimate_origin_asn] = [
                Ann(
                    prefix=CommonPrefixes.PREFIX.value,
                    as_path=(legitimate_origin_asn,),
                    next_hop_asn=legitimate_origin_asn,
                    recv_relationship=Relationships.ORIGIN,
                    timestamp=Timestamps.LEGITIMATE_ORIGIN,
                )
            ]
        seed_asn_ann_dict = {
            **seed_asn_ann_dict,
            **self._get_first_asn_stripped_attacker_seed_asn_ann_dict(engine),
        }
        return seed_asn_ann_dict

    def _get_first_asn_stripped_attacker_seed_asn_ann_dict(
        self, engine: SimulationEngine
    ) -> dict[int, list[Ann]]:
        attacker_seed_asn_ann_dict = self._get_attacker_seed_asn_ann_dict(engine)
        stripped_attacker_seed_asn_ann_dict: dict[int, list[Ann]] = {
            attacker_asn: [] for attacker_asn in self.attacker_asns
        }
        for attacker_asn, anns in attacker_seed_asn_ann_dict.items():
            for ann in anns:
                # Remove the attacker's ASN
                if len(ann.as_path) > 1:
                    stripped_ann = ann.copy(as_path=ann.as_path[1:])
                # Can't remove when path is less than or equal to 1
                else:
                    stripped_ann = ann
                stripped_attacker_seed_asn_ann_dict[attacker_asn].append(stripped_ann)
        return stripped_attacker_seed_asn_ann_dict

    def _get_roas(
        self,
        *,
        seed_asn_ann_dict: dict[int, list[Ann]],
        engine: SimulationEngine,
    ) -> list[ROA]:
        """Returns a list of ROAs"""

        return [
            ROA(CommonPrefixes.PREFIX.value, x) for x in self.legitimate_origin_asns
        ]

    def _get_dest_ip_addr(self) -> IPAddr:
        """Returns the destination IP address for the scenario"""

        return IPAddr("1.2.3.4")



from typing import Any

from bgpsimulator.shared import Prefix, Relationships


class Announcement:
    """BGP Announcement"""

    __slots__ = (
        "prefix",
        "as_path",
        "next_hop_asn",
        "recv_relationship",
        "timestamp",
        "bgpsec_next_asn",
        "bgpsec_as_path",
        "only_to_customers",
        "rovpp_blackhole",
        "rost_ids",
        "withdraw",
    )

    def __init__(
        self,
        prefix: Prefix,
        as_path: tuple[int, ...],
        next_hop_asn: int | None = None,
        recv_relationship: Relationships = Relationships.ORIGIN,
        timestamp: int = 0,
        bgpsec_next_asn: int | None = None,
        bgpsec_as_path: tuple[int, ...] | None = None,
        only_to_customers: int | None = None,
        rovpp_blackhole: bool = False,
        rost_ids: tuple[int, ...] | None = None,
        withdraw: bool = False,
    ):
        self.prefix: Prefix = prefix
        self.as_path: tuple[int, ...] = as_path
        # Equivalent to the next hop in a normal BGP announcement
        self.next_hop_asn: int = next_hop_asn or as_path[0]
        self.recv_relationship: Relationships = recv_relationship
        self.timestamp: int = timestamp
        self.bgpsec_next_asn: int | None = bgpsec_next_asn
        self.bgpsec_as_path: tuple[int, ...] = bgpsec_as_path or ()
        self.only_to_customers: int | None = only_to_customers
        self.rovpp_blackhole: bool = rovpp_blackhole
        self.rost_ids: tuple[int, ...] = rost_ids or ()
        self.withdraw: bool = withdraw
        if self.next_hop_asn is None:
            # next hop defaults to None, messing up the type
            if len(self.as_path) == 1:  # type: ignore
                self.next_hop_asn = self.as_path[0]
            elif len(self.as_path) > 1:
                raise ValueError(
                    "Announcement was initialized with an AS path longer than 1 "
                    f"({self.as_path}) but the next_hop_asn is ambiguous. "
                    " next_hop_asn is where the traffic should route to next."
                    "Please add "
                    "the next_hop_asn to the initialization parameters "
                    f"for the announcement of prefix {self.prefix}"
                )
            else:
                # Path is either zero or some other case we didn't account for
                raise NotImplementedError

    def copy(
        self,
        prefix: Prefix | None = None,
        as_path: tuple[int, ...] | None = None,
        next_hop_asn: int | None = None,
        recv_relationship: Relationships | None = None,
        timestamp: int | None = None,
        bgpsec_next_asn: int | None = None,
        bgpsec_as_path: tuple[int, ...] | None = None,
        only_to_customers: int | None = None,
        rovpp_blackhole: bool | None = None,
        rost_ids: tuple[int, ...] | None = None,
        withdraw: bool | None = None,
    ) -> "Announcement":
        """Creates a new announcement with the same attributes"""
        return Announcement(
            # NOTE: CANT USE OR!!! some of the actual vals are falsey
            prefix=prefix if prefix is not None else self.prefix,
            as_path=as_path if as_path is not None else self.as_path,
            next_hop_asn=next_hop_asn
            if next_hop_asn is not None
            else self.next_hop_asn,
            recv_relationship=recv_relationship
            if recv_relationship is not None
            else self.recv_relationship,
            timestamp=timestamp if timestamp is not None else self.timestamp,
            bgpsec_next_asn=bgpsec_next_asn
            if bgpsec_next_asn is not None
            else self.bgpsec_next_asn,
            bgpsec_as_path=bgpsec_as_path
            if bgpsec_as_path is not None
            else self.bgpsec_as_path,
            only_to_customers=only_to_customers
            if only_to_customers is not None
            else self.only_to_customers,
            rovpp_blackhole=rovpp_blackhole
            if rovpp_blackhole is not None
            else self.rovpp_blackhole,
            rost_ids=rost_ids if rost_ids is not None else self.rost_ids,
            withdraw=withdraw if withdraw is not None else self.withdraw,
        )

    def __repr__(self) -> str:
        return f"{self.prefix} {self.as_path} {self.recv_relationship.name}"

    def __eq__(self, other) -> bool:
        """Compare announcements by their attributes"""
        if not isinstance(other, Announcement):
            return NotImplemented
        return (
            self.prefix == other.prefix
            and self.as_path == other.as_path
            and self.next_hop_asn == other.next_hop_asn
            and self.recv_relationship == other.recv_relationship
            and self.timestamp == other.timestamp
            and self.bgpsec_next_asn == other.bgpsec_next_asn
            and self.bgpsec_as_path == other.bgpsec_as_path
            and self.only_to_customers == other.only_to_customers
            and self.rovpp_blackhole == other.rovpp_blackhole
            and self.rost_ids == other.rost_ids
            and self.withdraw == other.withdraw
        )

    @property
    def origin(self) -> int:
        """Returns the origin of the announcement"""

        return self.as_path[-1]

    def to_json(self) -> dict[str, Any]:
        """Converts the announcement to a JSON object"""
        return {
            "prefix": str(self.prefix),
            "as_path": list(self.as_path),
            "next_hop_asn": self.next_hop_asn,
            "recv_relationship": self.recv_relationship,
            "timestamp": self.timestamp,
            "bgpsec_next_asn": self.bgpsec_next_asn,
            "bgpsec_as_path": list(self.bgpsec_as_path),
            "only_to_customers": self.only_to_customers,
            "rovpp_blackhole": self.rovpp_blackhole,
            "rost_ids": list(self.rost_ids),
            "withdraw": self.withdraw,
        }

    @classmethod
    def from_json(cls, json_obj: dict[str, Any]) -> "Announcement":
        """Sensible defaults help with the website"""
        return cls(
            prefix=Prefix(json_obj["prefix"]),
            as_path=tuple(json_obj["as_path"]),
            next_hop_asn=json_obj.get("next_hop_asn", json_obj["as_path"][0]),
            recv_relationship=Relationships(json_obj["recv_relationship"]),
            timestamp=json_obj.get("timestamp", 0),
            bgpsec_next_asn=json_obj.get("bgpsec_next_asn"),
            bgpsec_as_path=tuple(json_obj.get("bgpsec_as_path", [])),
            only_to_customers=json_obj.get("only_to_customers", False),
            rovpp_blackhole=json_obj.get("rovpp_blackhole", False),
            rost_ids=tuple(json_obj.get("rost_ids", [])),
            withdraw=json_obj.get("withdraw", False),
        )



"""Shares chache dir, logger"""

import logging
from datetime import date
from pathlib import Path

from platformdirs import PlatformDirs, PlatformDirsABC

#############
# Cache Dir #
#############

# NOTE: Can't use getpass here due to windows bug (https://bugs.python.org/issue32731)
DIRS: PlatformDirsABC = PlatformDirs("bgpsimulator", Path.home().name)

SINGLE_DAY_CACHE_DIR: Path = Path(DIRS.user_cache_dir) / str(date.today())
SINGLE_DAY_CACHE_DIR.mkdir(exist_ok=True, parents=True)

##########
# Logger #
##########

bgpsimulator_logger = logging.getLogger("bgpsimulator")
bgpsimulator_logger.setLevel(logging.INFO)

# Add StreamHandler to output logs to console
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)

# Optional: Add a formatter for better readability
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
stream_handler.setFormatter(formatter)

# Attach handler to the logger
bgpsimulator_logger.addHandler(stream_handler)



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann


@dataclass(frozen=True)
class PolicyPropagateInfo:
    """Information about the policy propagation"""

    policy_propagate_bool: bool
    ann: "Ann"
    # Whether or not the announcement should be sent
    # Ex: In ROV++v1, the announcement is not sent if it's a blackhole,
    # but policy_propagate would still be true since the policy dealt with
    # propagation
    send_ann_bool: bool



from bgpsimulator.shared import Prefix

from .roa import ROA


class ROASNode:
    """Node in the ROAS tree"""

    def __init__(self, prefix: Prefix | None = None, roas: set[ROA] | None = None):
        self.prefix: Prefix | None = prefix
        self.roas: set[ROA] = roas or set()
        self.left: ROASNode | None = None
        self.right: ROASNode | None = None

    def add_data(self, prefix: Prefix, roa: ROA) -> None:
        """Adds ROA to the node for that prefix"""

        self.prefix = prefix
        self.roas.add(roa)



import json
from pathlib import Path

from bgpsimulator.simulation_framework.data_tracker.data_tracker import DataTracker
from bgpsimulator.simulation_framework.data_tracker.line_filter import LineFilter

from .line import Line
from .line_chart import LineChart


class LineChartFactory:
    """Factory for creating line charts"""

    def __init__(
        self,
        json_path: Path,
        graph_dir: Path,
        xlabel: str = "Percent Adoption",
        xlim: tuple[float, float] = (0, 100),
        ylim: tuple[float, float] = (0, 100),
        legend_loc: str = "best",
    ) -> None:
        self.data_tracker = DataTracker.from_json(json.loads(json_path.read_text()))
        self.graph_dir = graph_dir
        self.xlabel = xlabel
        self.xlim = xlim
        self.ylim = ylim
        self.legend_loc = legend_loc

    def generate_line_charts(self) -> None:
        """Generates graphs"""

        # Each line filter is a unique graph
        paths = self._generate_line_chart_jsons()
        paths = self._json_mod_hook(paths)
        self._write_pngs(paths)

    def _generate_line_chart_jsons(self) -> list[Path]:
        paths = []
        for line_filter in self.data_tracker.line_filters:
            paths.append(self._generate_line_chart_json(line_filter))
        return paths

    def _generate_line_chart_json(self, line_filter: LineFilter) -> Path:
        """Generates a graph for a given line filter and scenario label"""

        lines = []
        for scenario_label, inner_dict in self.data_tracker.aggregated_data.items():
            line = Line(scenario_label, [], [], [])
            for percent_ases_randomly_adopting, data_point in inner_dict[
                line_filter
            ].items():
                line.xs.append(percent_ases_randomly_adopting)
                line.ys.append(data_point["value"])
                line.yerrs.append(data_point["yerr"])
            lines.append(line)

        y_label = f"Percent {line_filter.outcome.name.replace('_', ' ').title()}"
        line_chart = LineChart(
            line_filter,
            lines,
            title=line_filter.to_csv(),
            xlabel=self.xlabel,
            ylabel=y_label,
            xlim=self.xlim,
            ylim=self.ylim,
            legend_loc=self.legend_loc,
        )
        line_chart_path = line_filter.get_json_path(self.graph_dir)
        line_chart_path.write_text(
            json.dumps(line_chart.to_json(), indent=4, sort_keys=True)
        )
        return line_chart_path

    def _json_mod_hook(self, paths: list[Path]) -> list[Path]:
        """Modifies the json files after they are generated"""
        return paths

    def _write_pngs(self, paths: list[Path]) -> None:
        """Writes the pngs"""
        for path in paths:
            line_chart = LineChart.from_json(json.loads(path.read_text()))
            png_path = line_chart.line_filter.get_png_path(self.graph_dir)
            line_chart.write_graph(png_path)



import argparse
import gc
import json
import os
import random
import shutil
import time
from copy import deepcopy
from functools import cached_property
from multiprocessing import Pool, cpu_count
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Any, Iterable
from warnings import warn

import psutil
from tqdm import tqdm

from bgpsimulator.as_graphs import (
    ASGraph,
    CAIDAASGraphCollector,
    CAIDAASGraphJSONConverter,
)
from bgpsimulator.shared import (
    ASNGroups,
    InAdoptingASNs,
    Outcomes,
    Settings,
    bgpsimulator_logger,
)
from bgpsimulator.simulation_engine import Policy, SimulationEngine

from .data_plane_packet_propagator import DataPlanePacketPropagator
from .data_tracker.data_tracker import DataTracker
from .data_tracker.line_filter import LineFilter
from .line_chart_factory.line_chart_factory import LineChartFactory
from .scenarios import SubprefixHijack
from .scenarios.scenario import Scenario
from .scenarios.scenario_config import ScenarioConfig

if TYPE_CHECKING:
    from multiprocessing.pool import ApplyResult

    from bgpsimulator.simulation_framework.scenarios.scenario import Scenario

parser = argparse.ArgumentParser(description="Runs BGPy simulations")
parser.add_argument(
    "--num_trials",
    "--trials",
    dest="trials",
    type=int,
    default=10,
    help="Number of trials to run",
)
parser.add_argument(
    "--parse_cpus",
    "--cpus",
    dest="cpus",
    type=int,
    default=max(cpu_count() - 1, 1),
    help="Number of CPUs to use for parsing",
)
# parse known args to avoid crashing during pytest
args, _unknown = parser.parse_known_args()


class Simulation:
    """A simulation of a BGP routing policy"""

    ##############
    # Init Funcs #
    ##############

    def __init__(
        self,
        output_dir: Path = Path("~/Desktop/sims").expanduser() / "bgpsimulator",
        percent_ases_randomly_adopting: tuple[float, ...] = (10, 20, 50, 80, 99),
        scenario_configs: tuple[ScenarioConfig, ...] = (
            ScenarioConfig(
                label="Subprefix Hijack; ROV Adopting",
                ScenarioCls=SubprefixHijack,
                default_adoption_settings={
                    Settings.ROV: True,
                },
            ),
        ),
        num_trials: int = args.trials,
        parse_cpus: int = args.cpus,
        python_hash_seed: int | None = None,
        as_graph_data_json_path: Path | None = None,
        line_filters: tuple[LineFilter, ...] = (),
    ) -> None:
        self.output_dir: Path = output_dir
        self.output_dir.mkdir(parents=True, exist_ok=True)
        self.percent_ases_randomly_adopting: tuple[float, ...] = (
            percent_ases_randomly_adopting
        )
        self.scenario_configs: tuple[ScenarioConfig, ...] = scenario_configs
        self.num_trials: int = num_trials
        self.parse_cpus: int = parse_cpus
        self.python_hash_seed: int | None = python_hash_seed
        self._seed_random()

        if not as_graph_data_json_path:
            caida_path: Path = CAIDAASGraphCollector().run()
            _, as_graph_data_json_path = CAIDAASGraphJSONConverter().run(
                caida_as_graph_path=caida_path
            )
        self.as_graph_data_json_path: Path = as_graph_data_json_path

        self.line_filters = line_filters
        if not self.line_filters:
            max_prop_rounds = max(x.propagation_rounds for x in self.scenario_configs)
            line_filters_list: list[LineFilter] = []
            for as_group in [ASNGroups.ALL_WOUT_IXPS]:
                for in_adopting_asns in InAdoptingASNs:
                    for outcome in Outcomes:
                        if outcome == Outcomes.UNDETERMINED:
                            continue
                        line_filters_list.append(
                            LineFilter(
                                as_group=as_group,
                                in_adopting_asns=in_adopting_asns,
                                # By default, we only track the last propagation round
                                prop_round=max_prop_rounds - 1,
                                outcome=outcome,
                            )
                        )
            self.line_filters = tuple(line_filters_list)

        # Can't delete this since it gets deleted in multiprocessing for some reason
        # NOTE: Once pypy gets to 3.12, just pass delete=False to this
        with TemporaryDirectory() as tmp_dir:
            tmp_dir_str = tmp_dir
        self._tqdm_tracking_dir: Path = Path(tmp_dir_str)
        self._tqdm_tracking_dir.mkdir(parents=True)

    def _seed_random(self, seed_suffix: str = "") -> None:
        """Seeds randomness"""

        if self.python_hash_seed is not None:
            msg = (
                f"You've set the python_hash_seed to {self.python_hash_seed}, but "
                "the simulations aren't deterministic unless you also set the "
                "PYTHONHASHSEED in the env, such as with \n"
                f"export PYTHONHASHSEED={self.python_hash_seed}"
            )
            if os.environ.get("PYTHONHASHSEED") != str(self.python_hash_seed):
                raise RuntimeError(msg)
            random.seed(str(self.python_hash_seed) + seed_suffix)

    def _validate_init(self):
        """Validates inputs to __init__

        Specifically checks for:
        1. scenario config mismatch between adopting and base policies
        2. duplicate scenario labels
        3. RAM constraints
        """

        self._validate_scenario_configs()
        self._validate_ram()

    def _validate_scenario_configs(self) -> None:
        """Validates ScenarioConfigs

        prevents duplicate scenario labels and ensures no mixups using BGPFull
        """

        scenario_labels = [x.label for x in self.scenario_configs]

        if len(set(scenario_labels)) != len(scenario_labels):
            raise ValueError(
                "Each ScenarioConfig uses a scenario_label when aggregating results "
                "Since you have two ScenarioConfig's with the same label, data "
                "won't be tracked properly. Please pass in a scenario_label= with a "
                "unique label name to your config"
            )

        for scenario_config in self.scenario_configs:
            if (
                scenario_config.num_attackers == 0
                and scenario_config.num_legitimate_origins == 0
            ):
                raise ValueError(
                    "ScenarioConfig has no attackers or victims. "
                    "Set num_attackers and/or num_legitimate_origins = 1 "
                    "or more "
                )

    def _validate_ram(self) -> None:
        """Validates that the RAM will not run out of bounds

        NOTE: all these values where obtained using pypy3.10 on a lenovo laptop
        """

        # NOTE: These are for PyPy, not Python
        total_gb_ram_per_core = 1.6

        expected_total_gb_ram = self.parse_cpus * total_gb_ram_per_core
        # Gets available RAM and converts to GB
        total_gb_ram = psutil.virtual_memory().available / (1024**3)

        bgpsimulator_logger.info(f"Expected RAM usage: {expected_total_gb_ram:.2f}")
        bgpsimulator_logger.info(f"Available RAM: {total_gb_ram:.2f}")
        if expected_total_gb_ram * 1.1 > total_gb_ram:
            warn(
                f"Estimated RAM usage is {expected_total_gb_ram:.2f}GB "
                f"but your machine has only {total_gb_ram:.2f}GB available, "
                "maybe use less cores or don't store provider/customer cones?",
                stacklevel=2,
            )

    ##############
    # Run Funcs #
    ##############

    def run(
        self,
        GraphFactoryCls: type[LineChartFactory] = LineChartFactory,
        graph_factory_kwargs: dict[str, Any] | None = None,
    ):  # , GraphFactoryCls: type[GraphFactory] = GraphFactory,
        # graph_factory_kwargs: dict[str, Any] | None = None) -> None:
        """Runs the simulation and writes the data"""

        data_tracker = self._get_data()
        self._write_data(data_tracker)
        self._graph_data(GraphFactoryCls, graph_factory_kwargs)
        # This object holds a lot of memory, good to get rid of it
        del data_tracker
        gc.collect()
        shutil.rmtree(self._tqdm_tracking_dir)

    def _get_data(self) -> DataTracker:
        """Runs trials for graph and aggregates data"""

        # Single process
        if self.parse_cpus == 1:
            # Results are a list of lists of metric trackers that we then sum
            return sum(
                self._get_single_process_results(),
                start=DataTracker(
                    line_filters=self.line_filters,
                    scenario_labels=self.scenario_labels,
                    percent_ases_randomly_adopting=self.percent_ases_randomly_adopting,
                ),
            )
        # Multiprocess
        else:
            # Results are a list of lists of metric trackers that we then sum
            return sum(
                self._get_mp_results(),
                start=DataTracker(
                    line_filters=self.line_filters,
                    scenario_labels=self.scenario_labels,
                    percent_ases_randomly_adopting=self.percent_ases_randomly_adopting,
                ),
            )

    ###########################
    # Multiprocessing Methods #
    ###########################

    def _get_chunks(self, cpus: int) -> list[list[int]]:
        """Returns chunks of trial inputs based on number of CPUs running

        Not a generator since we need this for multiprocessing

        We also don't multiprocess one by one because the start up cost of
        each process is huge (since each process must generate it's own engine
        ) so we must divy up the work beforehand
        """

        trials_list = list(range(self.num_trials))
        return [trials_list[i::cpus] for i in range(cpus)]

    def _get_single_process_results(self) -> list[DataTracker]:
        """Get all results when using single processing"""

        return [self._run_chunk(i, x) for i, x in enumerate(self._get_chunks(1))]

    def _get_mp_results(self) -> list[DataTracker]:
        """Get results from multiprocessing

        Previously used starmap, but now we have tqdm
        """

        # Pool is much faster than ProcessPoolExecutor
        with Pool(self.parse_cpus) as p:
            # return p.starmap(self._run_chunk, enumerate(self._get_chunks(parse_cpus)))
            chunks = self._get_chunks(self.parse_cpus)
            desc = f"Simulating {self.output_dir.name}"
            total = (
                sum(len(x) for x in chunks)
                * len(self.percent_ases_randomly_adopting)
                * len(self.scenario_configs)
            )
            with tqdm(total=total, desc=desc) as pbar:
                tasks: list[ApplyResult[DataTracker]] = [
                    p.apply_async(self._run_chunk, x) for x in enumerate(chunks)
                ]
                completed: list[DataTracker] = []
                while tasks:
                    completed, tasks = self._get_completed_and_tasks(completed, tasks)
                    self._update_tqdm_progress_bar(pbar)
                    time.sleep(0.5)
        return completed

    def _get_completed_and_tasks(self, completed, tasks):
        """Moves completed tasks into completed"""
        new_tasks = list()
        for task in tasks:
            if task.ready():
                completed.append(task.get())
            else:
                new_tasks.append(task)
        return completed, new_tasks

    def _update_tqdm_progress_bar(self, pbar: tqdm) -> None:  # type: ignore
        """Updates tqdm progress bar"""

        total_completed = 0
        for file_path in self._tqdm_tracking_dir.iterdir():
            try:
                total_completed += int(file_path.read_text())
            # Can happen write when file is being written when it's empty
            except ValueError:
                pass
        pbar.n = total_completed
        pbar.refresh()

    ############################
    # Data Aggregation Methods #
    ############################

    def _run_chunk(self, chunk_id: int, trials: list[int]) -> DataTracker:
        """Runs a chunk of trial inputs"""

        # Must also seed randomness here since we don't want multiproc to be the same
        self._seed_random(seed_suffix=str(chunk_id))

        # ASGraph is not picklable, so we need to create it here
        engine = SimulationEngine(
            as_graph=ASGraph.from_json(
                json.loads(self.as_graph_data_json_path.read_text())
            )
        )

        data_tracker = DataTracker(
            line_filters=self.line_filters,
            scenario_labels=self.scenario_labels,
            percent_ases_randomly_adopting=self.percent_ases_randomly_adopting,
        )

        for trial_index, trial in self._get_run_chunk_iter(trials):
            # Use the same attacker victim pairs across all percent adoptions
            trial_attacker_asns = None
            trial_legitimate_origin_asns = None
            for percent_adopt_index, percent_adopt in enumerate(
                self.percent_ases_randomly_adopting
            ):
                # Use the same adopting asns across all scenarios configs
                adopting_asns = None
                for scenario_config_index, scenario_config in enumerate(
                    self.scenario_configs
                ):
                    # Create the scenario for this trial
                    scenario = scenario_config.ScenarioCls(
                        scenario_config=scenario_config,
                        percent_ases_randomly_adopting=percent_adopt,
                        engine=engine,
                        route_validator=Policy.route_validator,
                        attacker_asns=trial_attacker_asns,
                        legitimate_origin_asns=trial_legitimate_origin_asns,
                        adopting_asns=adopting_asns,
                    )

                    # Change AS Classes, seed announcements before propagation
                    scenario.setup_engine(engine)
                    # For each round of propagation run the engine
                    for propagation_round in range(scenario_config.propagation_rounds):
                        self._single_engine_run(
                            engine=engine,
                            percent_ases_randomly_adopting=percent_adopt,
                            trial=trial,
                            scenario=scenario,
                            propagation_round=propagation_round,
                            data_tracker=data_tracker,
                        )
                    # Use the same attacker victim pairs across all percent adoptions
                    if self.reuse_attacker_asns:
                        trial_attacker_asns = scenario.attacker_asns
                    if self.reuse_legitimate_origin_asns:
                        trial_legitimate_origin_asns = scenario.legitimate_origin_asns
                    # Use the same adopting ASEs across all scenarios configs
                    if self.reuse_adopting_asns:
                        adopting_asns = scenario.adopting_asns

                    # Used to track progress with tqdm - update after each
                    # scenario_config
                    total_completed = (
                        trial_index
                        * len(self.percent_ases_randomly_adopting)
                        * len(self.scenario_configs)
                        + percent_adopt_index * len(self.scenario_configs)
                        + scenario_config_index
                        + 1
                    )
                    self._write_tqdm_progress(chunk_id, total_completed)

        self._write_tqdm_progress(
            chunk_id,
            len(trials)
            * len(self.percent_ases_randomly_adopting)
            * len(self.scenario_configs),
        )

        return data_tracker

    @cached_property
    def reuse_attacker_asns(self) -> bool:
        """Reuse the same attacker ASes across all scenarios configs.

        If they're from the same ASN group.
        """
        num_attackers_set = {x.num_attackers for x in self.scenario_configs}
        attacker_asn_groups_set = {x.attacker_asn_group for x in self.scenario_configs}
        return len(num_attackers_set) == 1 and len(attacker_asn_groups_set) == 1

    @cached_property
    def reuse_legitimate_origin_asns(self) -> bool:
        """Reuse the same victim ASes across all scenarios configs.

        If they're from the same ASN group.
        """
        num_legitimate_origins_set = {
            x.num_legitimate_origins for x in self.scenario_configs
        }
        legitimate_origin_asn_groups_set = {
            x.legitimate_origin_asn_group for x in self.scenario_configs
        }
        return (
            len(num_legitimate_origins_set) == 1
            and len(legitimate_origin_asn_groups_set) == 1
        )

    @cached_property
    def reuse_adopting_asns(self) -> bool:
        adoption_asn_groups_set = {
            tuple(x.adoption_asn_groups) for x in self.scenario_configs
        }
        return len(adoption_asn_groups_set) == 1

    def _get_run_chunk_iter(self, trials: list[int]) -> Iterable[tuple[int, int]]:
        """Returns iterator for trials with or without progress bar

        If there's only 1 cpu, run the progress bar here, else we run it elsewhere
        """

        if self.parse_cpus == 1:
            return tqdm(
                enumerate(trials),
                total=len(trials),
                desc=f"Simulating {self.output_dir.name}",
            )
        else:
            return enumerate(trials)

    def _write_tqdm_progress(self, chunk_id: int, completed: int) -> None:
        """Writes total number of percent adoption trial pairs to file"""

        # If self.parse_cpus == 1, then no multiprocessing is used
        if self.parse_cpus > 1:
            with (self._tqdm_tracking_dir / f"{chunk_id}.txt").open("w") as f:
                f.write(str(completed))

    def _single_engine_run(
        self,
        *,
        engine: SimulationEngine,
        percent_ases_randomly_adopting: float,
        trial: int,
        scenario: "Scenario",
        propagation_round: int,
        data_tracker: DataTracker,
    ) -> None:
        """Single engine run"""

        # Run the engine
        engine.propagate(propagation_round=propagation_round, scenario=scenario)

        # Pre-aggregation Hook
        scenario.pre_aggregation_hook(
            engine=engine,
            percent_ases_randomly_adopting=percent_ases_randomly_adopting,
            trial=trial,
            propagation_round=propagation_round,
        )

        self._collect_engine_run_data(
            engine,
            percent_ases_randomly_adopting,
            trial,
            scenario,
            propagation_round,
            data_tracker,
        )

        # By default, this is a no op
        scenario.post_propagation_hook(
            engine=engine,
            percent_ases_randomly_adopting=percent_ases_randomly_adopting,
            trial=trial,
            propagation_round=propagation_round,
        )

    def _collect_engine_run_data(
        self,
        engine: SimulationEngine,
        percent_ases_randomly_adopting: float,
        trial: int,
        scenario: "Scenario",
        propagation_round: int,
        data_tracker: DataTracker,
    ) -> dict[int, Outcomes]:
        # Save all engine run info
        # The reason we aggregate info right now, instead of saving
        # the engine and doing it later, is because doing it all
        # in RAM is MUCH faster, and speed is important
        outcomes = DataPlanePacketPropagator().get_as_outcomes_for_data_plane_packet(
            dest_ip_addr=scenario.dest_ip_addr,
            simulation_engine=engine,
            legitimate_origin_asns=scenario.legitimate_origin_asns,
            attacker_asns=scenario.attacker_asns,
            scenario=scenario,
        )

        data_tracker.store_trial_data(
            engine=engine,
            scenario=scenario,
            propagation_round=propagation_round,
            # dict[int, int] == dict[int, outcomes] (int enum)
            asn_to_packet_outcome_dict=outcomes,
        )
        return outcomes

    #######################
    # Graph Writing Funcs #
    #######################

    def _write_data(self, data_tracker: DataTracker) -> None:
        """Writes data to file"""

        data_tracker.aggregate_data()

        with self.json_path.open("w") as f:
            json.dump(data_tracker.to_json(), f, indent=4, sort_keys=True)
        with self.csv_path.open("w") as f:
            f.write(data_tracker.to_csv())
        bgpsimulator_logger.info(f"Wrote data to {self.json_path} and {self.csv_path}")

    def _graph_data(
        self,
        GraphFactoryCls: type[LineChartFactory] | None = LineChartFactory,
        kwargs=None,
    ) -> None:
        """Generates some default graphs"""

        # This prevents problems if kwargs is reused more than once
        # outside of this file, since we modify it
        kwargs = dict() if kwargs is None else deepcopy(kwargs)
        # Set defaults for kwargs
        kwargs["json_path"] = kwargs.pop("json_path", self.json_path)
        kwargs["graph_dir"] = kwargs.pop("graph_dir", self.graph_output_dir)
        if GraphFactoryCls:
            GraphFactoryCls(**kwargs).generate_line_charts()
            bgpsimulator_logger.info(f"\nWrote graphs to {kwargs['graph_dir']}")

    ##############
    # Properties #
    ##############

    @cached_property
    def scenario_labels(self) -> tuple[str, ...]:
        return tuple([x.label for x in self.scenario_configs])

    @cached_property
    def graph_output_dir(self) -> Path:
        return self.output_dir / "graphs"

    @cached_property
    def json_path(self) -> Path:
        return self.output_dir / "data.json"

    @cached_property
    def csv_path(self) -> Path:
        return self.output_dir / "data.csv"
